/*
 * arch/ppc/platforms/mousse_setup.c
 * 
 * Board setup routines for the Vooha Mousse Platform.
 *
 * Author: John W. Linville
 *         linville@lvl7.com
 *
 * Copyright 2002 LVL7 Systems Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/config.h>
#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/reboot.h>
#include <linux/pci.h>
#include <linux/kdev_t.h>
#include <linux/types.h>
#include <linux/major.h>
#include <linux/blk.h>
#include <linux/delay.h>
#include <linux/irq.h>
#include <linux/seq_file.h>

#include <asm/system.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <asm/time.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <asm/machdep.h>
#include <asm/prom.h>
#include <asm/open_pic.h>
#include <asm/todc.h>
#include <asm/bootinfo.h>
#include <asm/mpc10x.h>
#include <asm/pci-bridge.h>
#include <asm/ppcboot.h>

#include "mousse.h"

unsigned char __res[sizeof(bd_t)];

/*
 * *** IMPORTANT ***
 *
 * The first 16 entries of 'mousse_openpic_initsenses[]' are there and
 * initialized to 0 on purpose.  DO NOT REMOVE THEM as the 'offset' parameter
 * of 'openpic_init()' does not work for the mousse because the 8259
 * interrupt is NOT routed to the EPIC's IRQ 0 AND the EPIC's IRQ 0's offset is
 * the same as a normal openpic's IRQ 16 offset.
 */
static u_char mousse_openpic_initsenses[] __initdata = {
	0,	/* 0-15 not used by EPCI but by 8259 (std PC-type IRQs) */
	0,	/* 1 */
	0,	/* 2 */
	0,	/* 3 */
	0,	/* 4 */
	0,	/* 5 */
	0,	/* 6 */
	0,	/* 7 */
	0,	/* 8 */
	0,	/* 9 */
	0,	/* 10 */
	0,	/* 11 */
	0,	/* 12 */
	0,	/* 13 */
	0,	/* 14 */
	0,	/* 15 */
	0,	/* 16: EPIC IRQ 0: Active Low -- unused */
	1,	/* 17: EPIC IRQ 1: Active Low -- DEC 21143 */
	1,	/* 18: EPIC IRQ 2: Active Low -- Local PCI */
	1,	/* 19: EPIC IRQ 3: Active Low -- CPCI */
	1	/* 20: EPIC IRQ 4: Active Low -- Serial */
};

static void __init
mousse_setup_arch(void)
{
	loops_per_jiffy = 500000000 / HZ;

#ifdef CONFIG_BLK_DEV_INITRD
	if (initrd_start)
		ROOT_DEV = MKDEV(RAMDISK_MAJOR, 0);
	else
#endif
#ifdef	CONFIG_ROOT_NFS
		ROOT_DEV = to_kdev_t(0x00FF);	/* /dev/nfs pseudo device */
#else
		ROOT_DEV = to_kdev_t(0x0301);	/* /dev/hda1 IDE disk */
#endif

	/* Lookup PCI host bridges */
	mousse_find_bridges();

	printk("Vooha Mousse\n");
	printk("Mousse port (C) 2002 LVL7 Systems, Inc. (linville@lvl7.com)\n");
}

static ulong __init
mousse_find_end_of_memory(void)
{
	bd_t *binfo;

	binfo = (bd_t *)__res;

	return binfo->bi_memsize;
}

static void __init
mousse_map_io(void)
{
	io_block_mapping(0xff000000, 0xff000000, 0x01000000, _PAGE_IO);
}

static void __init
mousse_init_IRQ(void)
{
	OpenPIC_InitSenses = mousse_openpic_initsenses;
	OpenPIC_NumInitSenses = sizeof(mousse_openpic_initsenses);

	openpic_init(1, 0, NULL, -1);

	return;
}

static void
mousse_restart(char *cmd)
{
	__cli();

/* JWL -- use #define for base address... */
	for (;;) *((char *)0xffe09000) = 0; /* Spin until reset happens */

	panic("Restart failed.\n");
	/* NOTREACHED */
}

static void
mousse_power_off(void)
{
	__cli();

	for(;;);  /* No way to shut power off with software */

	panic("Power-off failed.\n");
	/* NOTREACHED */
}

static void
mousse_halt(void)
{
	mousse_power_off();
	/* NOTREACHED */
}

static int
mousse_show_cpuinfo(struct seq_file *m)
{
	bd_t *bp;
	uint pvid;

	bp = (bd_t *)__res;

	pvid = mfspr(PVR);

	seq_printf(m, "vendor\t\t: Vooha\n");
	seq_printf(m, "machine\t\t: Mousse\n");
	seq_printf(m, "processor\t: PVID: 0x%x, vendor: %s\n",
			pvid, (pvid & (1<<15) ? "IBM" : "Motorola"));
	seq_printf(m, "cpu speed\t: %ldMhz\n",
			bp->bi_intfreq/1000000);
	seq_printf(m, "bus speed\t: %ldMhz\n",
			bp->bi_busfreq/1000000);

	return 0;
}

/*
 * Set BAT 3 to map 0xfc000000 to end of physical memory space 1-to-1.
 */
static __inline__ void
mousse_set_bat(void)
{
	unsigned long   bat3u, bat3l;
	static int	mapping_set = 0;

	if (!mapping_set) {

		__asm__ __volatile__(
		" lis %0,0xfc00\n \
		  ori %1,%0,0x002a\n \
		  ori %0,%0,0x07fe\n \
		  mtspr 0x21e,%0\n \
		  mtspr 0x21f,%1\n \
		  isync\n \
		  sync "
		  : "=r" (bat3u), "=r" (bat3l));

		mapping_set = 1;
	}

	return;
}

#ifdef	CONFIG_SERIAL_TEXT_DEBUG
#include <linux/serialP.h>
#include <linux/serial_reg.h>
#include <asm/serial.h>

static struct serial_state rs_table[RS_TABLE_SIZE] = {
	SERIAL_PORT_DFNS	/* Defined in <asm/serial.h> */
};

static void
mousse_progress(char *s, unsigned short hex)
{
	volatile char c;
	volatile unsigned long com_port;
	u16 shift;

	com_port = rs_table[0].port;
	shift = rs_table[0].iomem_reg_shift;

	while ((c = *s++) != 0) {
		while ((*((volatile unsigned char *)com_port +
				(UART_LSR << shift)) & UART_LSR_THRE) == 0)
		                ;
	        *(volatile unsigned char *)com_port = c;
		
		if (c == '\n') {
			while ((*((volatile unsigned char *)com_port +
				(UART_LSR << shift)) & UART_LSR_THRE) == 0)
					;
	        	*(volatile unsigned char *)com_port = '\r';
		}
	}
}
#endif	/* CONFIG_SERIAL_TEXT_DEBUG */

void
mousse_get_ethaddr(unsigned char *ethaddr)
{
	bd_t *binfo = (bd_t *)__res;

	memcpy((void *)ethaddr, (void *)binfo->bi_enetaddr,
	       sizeof(binfo->bi_enetaddr));
}

unsigned char
nvram_read_byte(int i)
{
	if (ppc_md.nvram_read_val)
		return ppc_md.nvram_read_val(i);

        return 0;
}

void
nvram_write_byte(unsigned char c, int i)
{
	if (ppc_md.nvram_write_val)
		return ppc_md.nvram_write_val(i, c);

        return;
}

TODC_ALLOC();

void __init
platform_init(unsigned long r3, unsigned long r4, unsigned long r5,
	      unsigned long r6, unsigned long r7)
{
	bd_t *binfo;

	parse_bootinfo(find_bootinfo());

	if ( r3 )
		memcpy( (void *)__res,(void *)(r3+KERNELBASE), sizeof(bd_t) );

	binfo = (bd_t *)__res;

#ifdef CONFIG_BLK_DEV_INITRD
	/* take care of initrd if we have one */
	if ( r4 )
	{
		initrd_start = r4 + KERNELBASE;
		initrd_end = r5 + KERNELBASE;
	}
#endif /* CONFIG_BLK_DEV_INITRD */

	/* take care of cmd line */
	if ( r6 )
	{
		*(char *)(r7+KERNELBASE) = 0;

		/* nothing but serial consoles... */
		if (!strstr((char *)(r6+KERNELBASE), "console="))
			sprintf(cmd_line, "%s console=ttyS0,%d",
				(char *)(r6+KERNELBASE),
			        binfo->bi_baudrate);
		else
			strcpy(cmd_line, (char *)(r6+KERNELBASE));
	}

	/* Map in board regs, etc. */
	mousse_set_bat();

	isa_io_base = MPC10X_MAPB_ISA_IO_BASE;
	isa_mem_base = MPC10X_MAPB_ISA_MEM_BASE;
	pci_dram_offset = MPC10X_MAPB_DRAM_OFFSET;
	ISA_DMA_THRESHOLD = 0x00ffffff;
	DMA_MODE_READ = 0x44;
	DMA_MODE_WRITE = 0x48;

	ppc_md.setup_arch = mousse_setup_arch;
	ppc_md.show_cpuinfo = mousse_show_cpuinfo;
	ppc_md.irq_cannonicalize = NULL;
	ppc_md.init_IRQ = mousse_init_IRQ;
	ppc_md.get_irq = openpic_get_irq;
	ppc_md.init = NULL;

	ppc_md.restart = mousse_restart;
	ppc_md.power_off = mousse_power_off;
	ppc_md.halt = mousse_halt;

	ppc_md.find_end_of_memory = mousse_find_end_of_memory;
	ppc_md.setup_io_mappings = mousse_map_io;

/* JWL -- use #define for base address... */
	TODC_INIT(TODC_TYPE_MK48T59, 0, 0, 0xffe0a000, 0);
	ppc_md.time_init = todc_time_init;
	ppc_md.set_rtc_time = todc_set_rtc_time;
	ppc_md.get_rtc_time = todc_get_rtc_time;
	ppc_md.calibrate_decr = todc_calibrate_decr;

	ppc_md.nvram_read_val = todc_direct_read_val;
	ppc_md.nvram_write_val = todc_direct_write_val;

	ppc_md.heartbeat = NULL;
	ppc_md.heartbeat_reset = 0;
	ppc_md.heartbeat_count = 0;

#ifdef	CONFIG_SERIAL_TEXT_DEBUG
	ppc_md.progress = mousse_progress;
#else	/* !CONFIG_SERIAL_TEXT_DEBUG */
	ppc_md.progress = NULL;
#endif	/* CONFIG_SERIAL_TEXT_DEBUG */

	return;
}
