
/*
 * MTD driver for the 28F160C3 Flash Memory (non-CFI) on AmberJack.
 *
 * $Id: amberjack.c,v 1.1 2003/03/11 22:31:10 linville Exp $
 *
 * Author: John W. Linville <linville@lvl7.com>
 *
 * Copyright (c) 2002, LVL7, Inc.
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * References:
 *
 *    [1] 3 Volt Fast Boot Block Flash Memory" Intel Datasheet
 *           - Order Number: 290644-005
 *           - January 2000
 *
 *    [2] MTD internal API documentation
 *           - http://www.linux-mtd.infradead.org/tech/
 *
 * Limitations:
 *
 *    Even though this driver is written for 3 Volt Fast Boot
 *    Block Flash Memory, it is rather specific to AmberJack. With
 *    Minor modifications, notably the without data/address line
 *    mangling and different bus settings, etc. it should be
 *    trivial to adapt to other platforms.
 *
 */

/* debugging */
#undef AMBERJACK_DEBUG

/* partition support */
#define HAVE_PARTITIONS

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/mtd/mtd.h>
#ifdef HAVE_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

#ifndef CONFIG_LVL7_AJ
#error "This is for LVL7's AmberJack only!"
#endif

static char module_name[] = "amberjack";

/*
 * These values is specific to 28Fxxxx3 flash memory.
 * See section 2.3.1 in "3 Volt Fast Boot Block Flash Memory" Intel Datasheet
 */
#define FLASH_BLOCKSIZE_PARAM		(4096 * BUSWIDTH)
#define FLASH_NUMBLOCKS_16m_PARAM	8
#define FLASH_NUMBLOCKS_8m_PARAM	8

/*
 * These values is specific to 28Fxxxx3 flash memory.
 * See section 2.3.2 in "3 Volt Fast Boot Block Flash Memory" Intel Datasheet
 */
#define FLASH_BLOCKSIZE_MAIN		(32768 * BUSWIDTH)
#define FLASH_NUMBLOCKS_16m_MAIN	31
#define FLASH_NUMBLOCKS_8m_MAIN		15

/*
 * These values are specific to AmberJack
 */

/* general */
#define BUSWIDTH		2		/* don't change this - a lot of the code _will_ break if you change this */
#define FLASH_OFFSET		(0xff800000)

/* initial ramdisk */
#define INITRD_START		(0x00000000)
#define INITRD_LEN		(0x00200000)

/* app filesystem */
#define APPFS_START		(0x00200000)
#define APPFS_LEN		(0x00500000)

/* ppcboot */
#define PPCBOOT_START		(0x00700000)
#define PPCBOOT_LEN		(0x00060000)

/* kernel */
#define KERNEL_START		(0x00760000)
#define KERNEL_LEN		(0x00090000)

/* diagnostics */
#define DIAGS_START		(0x007F0000)
#define DIAGS_LEN		(0x00010000)

/*
 * See section 4.0 in "3 Volt Fast Boot Block Flash Memory" Intel Datasheet
 */
#define READ_ARRAY		0x00FF		/* Read Array/Reset */
#define READ_ID_CODES		0x0090		/* Read Identifier Codes */
#define ERASE_SETUP		0x0020		/* Block Erase */
#define ERASE_CONFIRM		0x00D0		/* Block Erase and Program Resume */
#define PGM_SETUP		0x0040		/* Program */
#define STATUS_READ		0x0070		/* Read Status Register */
#define STATUS_CLEAR		0x0050		/* Clear Status Register */
#define STATUS_BUSY		0x0080		/* Write State Machine Status (WSMS) */
#define STATUS_ERASE_ERR	0x0020		/* Erase Status (ES) */
#define STATUS_PGM_ERR		0x0010		/* Program Status (PS) */

/*
 * See section 4.2 in "3 Volt Fast Boot Block Flash Memory" Intel Datasheet
 */
#define FLASH_MANUFACTURER		0x0089
#define FLASH_DEVICE_8mbit_TOP		0x88c0
#define FLASH_DEVICE_8mbit_BOTTOM	0x88c1
#define FLASH_DEVICE_16mbit_TOP		0x88c2
#define FLASH_DEVICE_16mbit_BOTTOM	0x88c3

/***************************************************************************************************/

static struct mtd_info mtd;

static struct mtd_erase_region_info top_erase_regions[] =
{
   /* main blocks */
   {
	     offset: 0x00000000,
	  erasesize: FLASH_BLOCKSIZE_MAIN,
	  numblocks: FLASH_NUMBLOCKS_16m_MAIN
   },
   /* parameter blocks */
   {
	     offset: 0x00000000 + FLASH_BLOCKSIZE_MAIN * FLASH_NUMBLOCKS_16m_MAIN,
	  erasesize: FLASH_BLOCKSIZE_PARAM,
	  numblocks: FLASH_NUMBLOCKS_16m_PARAM
   }
};

static struct mtd_erase_region_info bottom_erase_regions[] =
{
   /* parameter blocks */
   {
	     offset: 0x00000000,
	  erasesize: FLASH_BLOCKSIZE_PARAM,
	  numblocks: FLASH_NUMBLOCKS_16m_PARAM
   },
   /* main blocks */
   {
	     offset: 0x00000000 + FLASH_BLOCKSIZE_PARAM * FLASH_NUMBLOCKS_16m_PARAM,
	  erasesize: FLASH_BLOCKSIZE_MAIN,
	  numblocks: FLASH_NUMBLOCKS_16m_MAIN
   }
};

static struct mtd_erase_region_info erase_regions[8];

#ifdef HAVE_PARTITIONS
static struct mtd_partition amberjack_partitions[] =
{
   /* initial ramdisk / file system */
   {
	       name: "root file system",
	     offset: INITRD_START,
	       size: INITRD_LEN,
	 mask_flags: 0
   },
   /* application file system */
   {
	       name: "app. file system",
	     offset: APPFS_START,
	       size: APPFS_LEN,
	 mask_flags: 0
   },
   /* ppcboot */
   {
	       name: "ppcboot",
	     offset: PPCBOOT_START,
	       size: PPCBOOT_LEN,
	 mask_flags: 0
   },
   /* kernel */
   {
	       name: "kernel",
	     offset: KERNEL_START,
	       size: KERNEL_LEN,
	 mask_flags: 0
   },
   /* diagnostics */
   {
	       name: "diagnostics",
	     offset: DIAGS_START,
	       size: DIAGS_LEN,
	 mask_flags: 0
   }
};
#endif

/***************************************************************************************************/

static __u8 read8 (__u32 offset)
{
   volatile __u8 *data = (__u8 *) (FLASH_OFFSET + offset);
#ifdef AMBERJACK_DEBUG
   printk (KERN_DEBUG "%s(): 0x%.8x -> 0x%.2x\n",__FUNCTION__,offset,*data);
#endif
   return (*data);
}

static __u16 read16 (__u32 offset)
{
   volatile __u16 *data = (__u16 *) (FLASH_OFFSET + offset);
#ifdef AMBERJACK_DEBUG
   printk (KERN_DEBUG "%s(): 0x%.8x -> 0x%.8x\n",__FUNCTION__,offset,*data);
#endif
   return (*data);
}

static void write16 (__u16 x,__u32 offset)
{
   volatile __u16 *data = (__u16 *) (FLASH_OFFSET + offset);
   *data = x;
#ifdef AMBERJACK_DEBUG
   printk (KERN_DEBUG "%s(): 0x%.8x <- 0x%.8x\n",__FUNCTION__,offset,*data);
#endif
}

/***************************************************************************************************/

/*
 * Probe for 16mbit flash memory on a AmberJack board.
 *
 * Returns 1 if we found 16mbit flash memory on AmberJack, 0 otherwise.
 */
static int flash_probe (void)
{
   int i;
   __u32 manufacturer,devtype;

   for (i=0; i<4; i++)
   {
      /* setup "Read Identifier Codes" mode */
      write16 (READ_ID_CODES, i*0x00200000);

      /* probe for ID */
      manufacturer = read16 (i*0x00200000);
      devtype = read16 ((i*0x00200000) + 2);

      /* put the flash back into command mode */
      write16 (READ_ARRAY, i*0x00200000);

      if (manufacturer != FLASH_MANUFACTURER)
         return (0);

      switch (devtype)
      {
         case FLASH_DEVICE_16mbit_TOP:
            memcpy(&(erase_regions[i*2]), top_erase_regions,
                   2*sizeof(struct mtd_erase_region_info));
            erase_regions[(i*2)].offset += (i*0x00200000);
            erase_regions[(i*2) + 1].offset += (i*0x00200000);
            break;

         case FLASH_DEVICE_16mbit_BOTTOM:
            memcpy(&(erase_regions[i*2]), bottom_erase_regions,
                   2*sizeof(struct mtd_erase_region_info));
            erase_regions[(i*2)].offset += (i*0x00200000);
            erase_regions[(i*2) + 1].offset += (i*0x00200000);
            break;

         default:
            return (0);
      }
   }

   return (1);
}

/*
 * Erase one block of flash memory at offset ``offset'' which is any
 * address within the block which should be erased.
 *
 * Returns 1 if successful, 0 otherwise.
 */
static inline int erase_block (__u32 offset)
{
   __u32 status;

#ifdef AMBERJACK_DEBUG
   printk (KERN_DEBUG "%s(): 0x%.8x\n",__FUNCTION__,offset);
#endif

   /* erase and confirm */
   write16 (ERASE_SETUP,offset);
   write16 (ERASE_CONFIRM,offset);

   /* wait for block erase to finish */
   do
	 {
		write16 (STATUS_READ,offset);
		status = read16 (offset);
	 }
   while ((~status & STATUS_BUSY) != 0);

   /* put the flash back into command mode */
   write16 (READ_ARRAY,offset);

   /* was the erase successfull? */
   if ((status & STATUS_ERASE_ERR))
	 {
		printk (KERN_WARNING "%s: erase error at address 0x%.8x.\n",module_name,offset);
		return (0);
	 }

   return (1);
}

static int flash_erase (struct mtd_info *mtd,struct erase_info *instr)
{
   __u32 addr,len;
   int i,first;

#ifdef AMBERJACK_DEBUG
   printk (KERN_DEBUG "%s(addr = 0x%.8x, len = %d)\n",__FUNCTION__,instr->addr,instr->len);
#endif

   /* sanity checks */
   if (instr->addr + instr->len > mtd->size) return (-EINVAL);

   /*
    * check that both start and end of the requested erase are
    * aligned with the erasesize at the appropriate addresses.
    *
    * skip all erase regions which are ended before the start of
    * the requested erase. Actually, to save on the calculations,
    * we skip to the first erase region which starts after the
    * start of the requested erase, and then go back one.
    */
   for (i = 0; i < mtd->numeraseregions && instr->addr >= mtd->eraseregions[i].offset; i++) ;
   i--;

   /*
    * ok, now i is pointing at the erase region in which this
    * erase request starts. Check the start of the requested
    * erase range is aligned with the erase size which is in
    * effect here.
    */
   if (instr->addr & (mtd->eraseregions[i].erasesize - 1)) return (-EINVAL);

   /* Remember the erase region we start on */
   first = i;

   /*
    * next, check that the end of the requested erase is aligned
    * with the erase region at that address.
    *
    * as before, drop back one to point at the region in which
    * the address actually falls
    */
   for (; i < mtd->numeraseregions && instr->addr + instr->len >= mtd->eraseregions[i].offset; i++) ;
   i--;

   /* is the end aligned on a block boundary? */
   if ((instr->addr + instr->len) & (mtd->eraseregions[i].erasesize - 1)) return (-EINVAL);

   addr = instr->addr;
   len = instr->len;

   i = first;

   /* now erase those blocks */
   while (len)
	 {
		if (!erase_block (addr))
		  {
			 instr->state = MTD_ERASE_FAILED;
			 return (-EIO);
		  }

		addr += mtd->eraseregions[i].erasesize;
		len -= mtd->eraseregions[i].erasesize;

		if (addr == mtd->eraseregions[i].offset + (mtd->eraseregions[i].erasesize * mtd->eraseregions[i].numblocks)) i++;
	 }

   instr->state = MTD_ERASE_DONE;
   if (instr->callback) instr->callback (instr);

   return (0);
}

static int flash_read (struct mtd_info *mtd,loff_t from,size_t len,size_t *retlen,u_char *buf)
{
#ifdef AMBERJACK_DEBUG
   printk (KERN_DEBUG "%s(from = 0x%.8x, len = %d)\n",__FUNCTION__,(__u32) from,len);
#endif

   /* sanity checks */
   if (!len) return (0);
   if (from + len > mtd->size) return (-EINVAL);

   /* we always read len bytes */
   *retlen = len;

   /* first, we read bytes until we reach a dword boundary */
   if (from & (BUSWIDTH - 1))
	 {
		int gap = BUSWIDTH - (from & (BUSWIDTH - 1));

		while (len && gap--) *buf++ = read8 (from++), len--;
	 }

   /* now we read dwords until we reach a non-dword boundary */
   while (len >= BUSWIDTH)
	 {
		*((__u16 *) buf) = read16 (from);

		buf += BUSWIDTH;
		from += BUSWIDTH;
		len -= BUSWIDTH;
	 }

   /* top up the last unaligned bytes */
   if (len & (BUSWIDTH - 1))
	 while (len--) *buf++ = read8 (from++);

   return (0);
}

/*
 * Write one dword ``x'' to flash memory at offset ``offset''. ``offset''
 * must be 32 bits, i.e. it must be on a dword boundary.
 *
 * Returns 1 if successful, 0 otherwise.
 */
static inline int write_word (__u32 offset,__u16 x)
{
   __u32 status;

#ifdef AMBERJACK_DEBUG
   printk (KERN_DEBUG "%s(): 0x%.8x <- 0x%.8x\n",__FUNCTION__,offset,x);
#endif

   /* setup writing */
   write16 (PGM_SETUP,offset);

   /* write the data */
   write16 (x,offset);

   /* wait for the write to finish */
   do
	 {
		write16 (STATUS_READ,offset);
		status = read16 (offset);
	 }
   while ((~status & STATUS_BUSY) != 0);

   /* put the flash back into command mode */
   write16 (READ_ARRAY,offset);

   /* was the write successfull? */
   if ((status & STATUS_PGM_ERR) || read16 (offset) != x)
	 {
		printk (KERN_WARNING "%s: write error at address 0x%.8x.\n",module_name,offset);
		return (0);
	 }

   return (1);
}

static int flash_write (struct mtd_info *mtd,loff_t to,size_t len,size_t *retlen,const u_char *buf)
{
   __u8 tmp[BUSWIDTH];
   int i,n;

#ifdef AMBERJACK_DEBUG
   printk (KERN_DEBUG "%s(to = 0x%.8x, len = %d)\n",__FUNCTION__,(__u32) to,len);
#endif

   *retlen = 0;

   /* sanity checks */
   if (!len) return (0);
   if (to + len > mtd->size) return (-EINVAL);

   /* first, we write a 0xFF.... padded byte until we reach a dword boundary */
   if (to & (BUSWIDTH - 1))
	 {
		__u32 aligned = to & ~(BUSWIDTH - 1);
		int gap = to - aligned;

		i = n = 0;

		while (gap--) tmp[i++] = 0xFF;
		while (len && i < BUSWIDTH) tmp[i++] = buf[n++], len--;
		while (i < BUSWIDTH) tmp[i++] = 0xFF;

		if (!write_word (aligned,*((__u16 *) tmp))) return (-EIO);

		to += n;
		buf += n;
		*retlen += n;
	 }

   /* now we write dwords until we reach a non-dword boundary */
   while (len >= BUSWIDTH)
	 {
		if (!write_word (to,*((__u16 *) buf))) return (-EIO);

		to += BUSWIDTH;
		buf += BUSWIDTH;
		*retlen += BUSWIDTH;
		len -= BUSWIDTH;
	 }

   /* top up the last unaligned bytes, padded with 0xFF.... */
   if (len & (BUSWIDTH - 1))
	 {
		i = n = 0;

		while (len--) tmp[i++] = buf[n++];
		while (i < BUSWIDTH) tmp[i++] = 0xFF;

		if (!write_word (to,*((__u16 *) tmp))) return (-EIO);

		*retlen += n;
	 }

   return (0);
}

static void flash_sync(struct mtd_info *mtd)
{

   /* mtdpart.c assumes that we have a sync function...  probably a bug,
      but easy enough to fix here instead... */

}

/***************************************************************************************************/

#define NB_OF(x) (sizeof (x) / sizeof (x[0]))

int __init amberjack_flash_init (void)
{
   int result;
   memset (&mtd,0,sizeof (mtd));
   printk ("MTD driver for AmberJack. Written by John W. Linville <linville@lvl7.com>\n");
   printk ("%s: Probing for 28F160C3x flash on AmberJack...\n",module_name);
   if (!flash_probe ())
	 {
		printk (KERN_WARNING "%s: Found no AmberJack compatible flash device\n",module_name);
		return (-ENXIO);
	 }
   printk ("%s: This looks like an AmberJack board to me.\n",module_name);
   mtd.name = module_name;
   mtd.type = MTD_NORFLASH;
   mtd.flags = MTD_CAP_NORFLASH;
   mtd.size = 4 * (FLASH_BLOCKSIZE_PARAM * FLASH_NUMBLOCKS_16m_PARAM
                 + FLASH_BLOCKSIZE_MAIN * FLASH_NUMBLOCKS_16m_MAIN);
   mtd.erasesize = FLASH_BLOCKSIZE_MAIN;
   mtd.numeraseregions = NB_OF (erase_regions);
   mtd.eraseregions = erase_regions;
   mtd.module = THIS_MODULE;
   mtd.erase = flash_erase;
   mtd.read = flash_read;
   mtd.write = flash_write;
   mtd.sync = flash_sync;

#ifdef AMBERJACK_DEBUG
   printk (KERN_DEBUG
		   "mtd.name = %s\n"
		   "mtd.size = 0x%.8x (%uM)\n"
		   "mtd.erasesize = 0x%.8x (%uK)\n"
		   "mtd.numeraseregions = %d\n",
		   mtd.name,
		   mtd.size,mtd.size / (1024*1024),
		   mtd.erasesize,mtd.erasesize / 1024,
		   mtd.numeraseregions);

   if (mtd.numeraseregions)
	 for (result = 0; result < mtd.numeraseregions; result++)
	   printk (KERN_DEBUG
			   "\n\n"
			   "mtd.eraseregions[%d].offset = 0x%.8x\n"
			   "mtd.eraseregions[%d].erasesize = 0x%.8x (%uK)\n"
			   "mtd.eraseregions[%d].numblocks = %d\n",
			   result,mtd.eraseregions[result].offset,
			   result,mtd.eraseregions[result].erasesize,mtd.eraseregions[result].erasesize / 1024,
			   result,mtd.eraseregions[result].numblocks);

#ifdef HAVE_PARTITIONS
   printk ("\npartitions = %d\n",NB_OF (amberjack_partitions));

   for (result = 0; result < NB_OF (amberjack_partitions); result++)
	 printk (KERN_DEBUG
			 "\n\n"
			 "amberjack_partitions[%d].name = %s\n"
			 "amberjack_partitions[%d].offset = 0x%.8x\n"
			 "amberjack_partitions[%d].size = 0x%.8x (%uK)\n",
			 result,amberjack_partitions[result].name,
			 result,amberjack_partitions[result].offset,
			 result,amberjack_partitions[result].size,amberjack_partitions[result].size / 1024);
#endif
#endif

#ifndef HAVE_PARTITIONS
   result = add_mtd_device (&mtd);
#else
   result = add_mtd_partitions (&mtd,amberjack_partitions,NB_OF (amberjack_partitions));
#endif

   return (result);
}

void __exit amberjack_flash_exit (void)
{
#ifndef HAVE_PARTITIONS
   del_mtd_device (&mtd);
#else
   del_mtd_partitions (&mtd);
#endif
}

module_init (amberjack_flash_init);
module_exit (amberjack_flash_exit);

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
MODULE_AUTHOR("John W. Linville <linville@lvl7.com>");
MODULE_DESCRIPTION("MTD driver for Intel 28F160C3x on AmberJack board");


