/*
 * MTD driver for the 28F160C3B Flash Memory (non-CFI) on Accton Apollo.
 *
 * $Id: apollo.c,v 1.1 2003/03/11 22:31:10 linville Exp $
 *
 * Author: John W. Linville <linville@lvl7.com>
 *
 * Copyright (c) 2002, LVL7 Systems, Inc.
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * References:
 *
 *    [1] 3 Volt Fast Boot Block Flash Memory" Intel Datasheet
 *           - Order Number: 290644-005
 *           - January 2000
 *
 *    [2] MTD internal API documentation
 *           - http://www.linux-mtd.infradead.org/tech/
 *
 * Limitations:
 *
 *    Even though this driver is written for 3 Volt Fast Boot
 *    Block Flash Memory, it is rather specific to Apollo. With
 *    Minor modifications, notably the without data/address line
 *    mangling and different bus settings, etc. it should be
 *    trivial to adapt to other platforms.
 *
 */

/* debugging */
#undef APOLLO_DEBUG

/* partition support */
#define HAVE_PARTITIONS

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/mtd/mtd.h>
#ifdef HAVE_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

#ifndef CONFIG_ACCTON_APOLLO
#error "This is for the Accton Apollo ONLY!"
#endif

static char module_name[] = "apollo";

/*
 * These values is specific to 28Fxxxx3 flash memory.
 * See section 2.3.1 in "3 Volt Fast Boot Block Flash Memory" Intel Datasheet
 */
#define FLASH_BLOCKSIZE_PARAM		(4096 * BUSWIDTH)
#define FLASH_NUMBLOCKS_16m_PARAM	8
#define FLASH_NUMBLOCKS_8m_PARAM	8

/*
 * These values is specific to 28Fxxxx3 flash memory.
 * See section 2.3.2 in "3 Volt Fast Boot Block Flash Memory" Intel Datasheet
 */
#define FLASH_BLOCKSIZE_MAIN		(32768 * BUSWIDTH)
#define FLASH_NUMBLOCKS_16m_MAIN	31
#define FLASH_NUMBLOCKS_8m_MAIN		15

/*
 * These values are specific to Apollo
 */

/* general */
#define BUSWIDTH		8		/* don't change this - a lot of the code _will_ break if you change this */
#define FLASH_OFFSET		(0xff800000)

/* PPCBoot environment */
#define PPCBOOT_ENV_START	(0x00000000)
#define PPCBOOT_ENV_LEN		(0x00040000)

/* initial ramdisk */
#define INITRD_START		(0x00040000)
#define INITRD_LEN		(0x000C0000)

/* root ramdisk */
#define ROOTFS_START		(0x00100000)
#define ROOTFS_LEN		(0x00180000)

/* app filesystem */
#define APPFS_START		(0x00280000)
#define APPFS_LEN		(0x00480000)

/* ppcboot */
#define PPCBOOT_START		(0x00700000)
#define PPCBOOT_LEN		(0x00040000)

/* kernel */
#define KERNEL_START		(0x00740000)
#define KERNEL_LEN		(0x000c0000)

/*
 * See section 4.0 in "3 Volt Fast Boot Block Flash Memory" Intel Datasheet
 */
#define READ_ARRAY		0x00FF00FF00FF00FF	/* Read Array/Reset */
#define READ_ID_CODES		0x0090009000900090	/* Read Identifier Codes */
#define ERASE_SETUP		0x0020002000200020	/* Block Erase */
#define ERASE_CONFIRM		0x00D000D000D000D0	/* Block Erase and Program Resume */
#define PGM_SETUP		0x0040004000400040	/* Program */
#define STATUS_READ		0x0070007000700070	/* Read Status Register */
#define STATUS_CLEAR		0x0050005000500050	/* Clear Status Register */
#define STATUS_BUSY		0x0080008000800080	/* Write State Machine Status (WSMS) */
#define STATUS_ERASE_ERR	0x0020002000200020	/* Erase Status (ES) */
#define STATUS_PGM_ERR		0x0010001000100010	/* Program Status (PS) */

/*
 * See section 4.2 in "3 Volt Fast Boot Block Flash Memory" Intel Datasheet
 */
#define FLASH_MANUFACTURER		0x0089008900890089

/*
 * Compiler doesn't like switch on __u64 type, so define these as __u16...
 */
#define FLASH_DEVICE_8mbit_TOP		0x88c0
#define FLASH_DEVICE_8mbit_BOTTOM	0x88c1
#define FLASH_DEVICE_16mbit_TOP		0x88c2
#define FLASH_DEVICE_16mbit_BOTTOM	0x88c3

/***************************************************************************************************/

static struct mtd_info mtd;

static struct mtd_erase_region_info top_erase_regions[] =
{
   /* main blocks */
   {
	     offset: 0x00000000,
	  erasesize: FLASH_BLOCKSIZE_MAIN,
	  numblocks: FLASH_NUMBLOCKS_16m_MAIN
   },
   /* parameter blocks */
   {
	     offset: 0x00000000 + FLASH_BLOCKSIZE_MAIN * FLASH_NUMBLOCKS_16m_MAIN,
	  erasesize: FLASH_BLOCKSIZE_PARAM,
	  numblocks: FLASH_NUMBLOCKS_16m_PARAM
   }
};

static struct mtd_erase_region_info bottom_erase_regions[] =
{
   /* parameter blocks */
   {
	     offset: 0x00000000,
	  erasesize: FLASH_BLOCKSIZE_PARAM,
	  numblocks: FLASH_NUMBLOCKS_16m_PARAM
   },
   /* main blocks */
   {
	     offset: 0x00000000 + FLASH_BLOCKSIZE_PARAM * FLASH_NUMBLOCKS_16m_PARAM,
	  erasesize: FLASH_BLOCKSIZE_MAIN,
	  numblocks: FLASH_NUMBLOCKS_16m_MAIN
   }
};

static struct mtd_erase_region_info erase_regions[2];

#ifdef HAVE_PARTITIONS
static struct mtd_partition apollo_partitions[] =
{
   /* PPCBoot environment */
   {
	       name: "PPCBoot Environment",
	     offset: PPCBOOT_ENV_START,
	       size: PPCBOOT_ENV_LEN,
	 mask_flags: 0
   },
   /* initial ramdisk / file system */
   {
	       name: "Initrd Image",
	     offset: INITRD_START,
	       size: INITRD_LEN,
	 mask_flags: 0
   },
   /* root ramdisk / file system */
   {
	       name: "Root FS Image",
	     offset: ROOTFS_START,
	       size: ROOTFS_LEN,
	 mask_flags: 0
   },
   /* application file system */
   {
	       name: "Application FS",
	     offset: APPFS_START,
	       size: APPFS_LEN,
	 mask_flags: 0
   },
   /* ppcboot */
   {
	       name: "PPCBoot",
	     offset: PPCBOOT_START,
	       size: PPCBOOT_LEN,
	 mask_flags: 0
   },
   /* kernel */
   {
	       name: "Linux Kernel Image",
	     offset: KERNEL_START,
	       size: KERNEL_LEN,
	 mask_flags: 0
   }
};
#endif

/***************************************************************************************************/

static __u8 read8 (__u32 offset)
{
   volatile __u8 *data = (__u8 *) (FLASH_OFFSET + offset);
#ifdef APOLLO_DEBUG
   printk (KERN_DEBUG "%s(): 0x%.8x -> 0x%.2x\n",__FUNCTION__,offset,*data);
#endif
   return (*data);
}

/* Access routines for Apollo's RTC -- 8 bits accessed as 64... */
static void read64 (__u32 offset, __u64 *data)
{
   __u32 flags, msr, tmp_msr;
   volatile __u64 save_reg, *save_ptr;
   volatile __u64 *addr = (__u64 *) (FLASH_OFFSET + offset);

   save_ptr = &save_reg;

   save_flags(flags);
   cli();

   __asm__ __volatile__ ("mfmsr %0" : "=r" (msr));

   tmp_msr = msr | MSR_FP;
   __asm__ __volatile__ ("mtmsr %0" : : "r" (tmp_msr));
   __asm__ __volatile__ ("eieio ; sync");

   __asm__ __volatile__ ("stfd  1, 0(%0)" : : "r" (save_ptr));
   __asm__ __volatile__ ("eieio ; sync");
   __asm__ __volatile__ ("lfd  1, 0(%0)" : : "r" (addr));
   __asm__ __volatile__ ("stfd  1, 0(%0)" : : "r" (data));
   __asm__ __volatile__ ("eieio ; sync");
   __asm__ __volatile__ ("lfd  1, 0(%0)" : : "r" (save_ptr));
   __asm__ __volatile__ ("eieio ; sync");

   __asm__ __volatile__ ("mtmsr %0" : : "r" (msr));

   restore_flags(flags);

#ifdef APOLLO_DEBUG
   printk (KERN_DEBUG "%s(): 0x%.8x -> 0x%.16x\n",__FUNCTION__,offset,*data);
#endif
}

static void write64 (__u32 offset, __u64 *data)
{
   __u32 flags, msr, tmp_msr;
   volatile __u64 save_reg, *save_ptr;
   volatile __u64 *addr = (__u64 *) (FLASH_OFFSET + offset);

#ifdef APOLLO_DEBUG
   printk (KERN_DEBUG "%s(): 0x%.8x -> 0x%.16x\n",__FUNCTION__,offset,*data);
#endif

   save_ptr = &save_reg;

   save_flags(flags);
   cli();

   __asm__ __volatile__ ("mfmsr %0" : "=r" (msr));

   tmp_msr = msr | MSR_FP;
   __asm__ __volatile__ ("mtmsr %0" : : "r" (tmp_msr));
   __asm__ __volatile__ ("eieio ; sync");

   __asm__ __volatile__ ("stfd  1, 0(%0)" : : "r" (save_ptr));
   __asm__ __volatile__ ("eieio ; sync");
   __asm__ __volatile__ ("lfd  1, 0(%0)" : : "r" (data));
   __asm__ __volatile__ ("stfd  1, 0(%0)" : : "r" (addr));
   __asm__ __volatile__ ("eieio ; sync");
   __asm__ __volatile__ ("lfd  1, 0(%0)" : : "r" (save_ptr));
   __asm__ __volatile__ ("eieio ; sync");

   __asm__ __volatile__ ("mtmsr %0" : : "r" (msr));

   restore_flags(flags);
}

/***************************************************************************************************/

/*
 * Probe for 16mbit flash memory on a Apollo board.
 *
 * Returns 1 if we found 16mbit flash memory on Apollo, 0 otherwise.
 */
static int flash_probe (void)
{
   __u64 command;
   __u64 manufacturer,devtype;

   /* setup "Read Identifier Codes" mode */
   command = READ_ID_CODES;
   write64 (0, &command);

   /* probe for ID */
   read64 (0, &manufacturer);
   read64 (8, &devtype);

   /* put the flash back into command mode */
   command = READ_ARRAY;
   write64 (0, &command);

   if (manufacturer != FLASH_MANUFACTURER)
      return (0);

   /* Compiler didn't like switch on __u64 type, so... */
   switch ((__u16)(devtype & 0x000000000000ffff))
   {
      case FLASH_DEVICE_16mbit_TOP:
         memcpy(&(erase_regions[0]), top_erase_regions,
                2*sizeof(struct mtd_erase_region_info));
         break;

      case FLASH_DEVICE_16mbit_BOTTOM:
         memcpy(&(erase_regions[0]), bottom_erase_regions,
                2*sizeof(struct mtd_erase_region_info));
         break;

      default:
         return (0);
   }

   return (1);
}

/*
 * Erase one block of flash memory at offset ``offset'' which is any
 * address within the block which should be erased.
 *
 * Returns 1 if successful, 0 otherwise.
 */
static inline int erase_block (__u32 offset)
{
   __u64 status, command;

#ifdef APOLLO_DEBUG
   printk (KERN_DEBUG "%s(): 0x%.8x\n",__FUNCTION__,offset);
#endif

   /* erase and confirm */
   command = ERASE_SETUP;
   write64 (offset, &command);
   command = ERASE_CONFIRM;
   write64 (offset, &command);

   /* wait for block erase to finish */
   do
	 {
		read64 (offset, &status);
	 }
   while ((~status & STATUS_BUSY) != 0);

   /* put the flash back into command mode */
   command = READ_ARRAY;
   write64 (offset, &command);

   /* was the erase successfull? */
   if ((status & STATUS_ERASE_ERR))
	 {
		printk (KERN_WARNING "%s: erase error at address 0x%.8x.\n",module_name,offset);
		return (0);
	 }

   return (1);
}

static int flash_erase (struct mtd_info *mtd,struct erase_info *instr)
{
   __u32 addr,len;
   int i,first;

#ifdef APOLLO_DEBUG
   printk (KERN_DEBUG "%s(addr = 0x%.8x, len = %d)\n",__FUNCTION__,instr->addr,instr->len);
#endif

   /* sanity checks */
   if (instr->addr + instr->len > mtd->size) return (-EINVAL);

   /*
    * check that both start and end of the requested erase are
    * aligned with the erasesize at the appropriate addresses.
    *
    * skip all erase regions which are ended before the start of
    * the requested erase. Actually, to save on the calculations,
    * we skip to the first erase region which starts after the
    * start of the requested erase, and then go back one.
    */
   for (i = 0; i < mtd->numeraseregions && instr->addr >= mtd->eraseregions[i].offset; i++) ;
   i--;

   /*
    * ok, now i is pointing at the erase region in which this
    * erase request starts. Check the start of the requested
    * erase range is aligned with the erase size which is in
    * effect here.
    */
   if (instr->addr & (mtd->eraseregions[i].erasesize - 1)) return (-EINVAL);

   /* Remember the erase region we start on */
   first = i;

   /*
    * next, check that the end of the requested erase is aligned
    * with the erase region at that address.
    *
    * as before, drop back one to point at the region in which
    * the address actually falls
    */
   for (; i < mtd->numeraseregions && instr->addr + instr->len >= mtd->eraseregions[i].offset; i++) ;
   i--;

   /* is the end aligned on a block boundary? */
   if ((instr->addr + instr->len) & (mtd->eraseregions[i].erasesize - 1)) return (-EINVAL);

   addr = instr->addr;
   len = instr->len;

   i = first;

   /* now erase those blocks */
   while (len)
	 {
		if (!erase_block (addr))
		  {
			 instr->state = MTD_ERASE_FAILED;
			 return (-EIO);
		  }

		addr += mtd->eraseregions[i].erasesize;
		len -= mtd->eraseregions[i].erasesize;

		if (addr == mtd->eraseregions[i].offset + (mtd->eraseregions[i].erasesize * mtd->eraseregions[i].numblocks)) i++;
	 }

   instr->state = MTD_ERASE_DONE;
   if (instr->callback) instr->callback (instr);

   return (0);
}

static int flash_read (struct mtd_info *mtd,loff_t from,size_t len,size_t *retlen,u_char *buf)
{
#ifdef APOLLO_DEBUG
   printk (KERN_DEBUG "%s(from = 0x%.8x, len = %d)\n",__FUNCTION__,(__u32) from,len);
#endif

   /* sanity checks */
   if (!len) return (0);
   if (from + len > mtd->size) return (-EINVAL);

   /* we always read len bytes */
   *retlen = len;

   /* first, we read bytes until we reach a dword boundary */
   if (from & (BUSWIDTH - 1))
	 {
		int gap = BUSWIDTH - (from & (BUSWIDTH - 1));

		while (len && gap--) *buf++ = read8 (from++), len--;
	 }

   /* now we read dwords until we reach a non-dword boundary */
   while (len >= BUSWIDTH)
	 {
		read64 (from, (__u64 *)buf);

		buf += BUSWIDTH;
		from += BUSWIDTH;
		len -= BUSWIDTH;
	 }

   /* top up the last unaligned bytes */
   if (len & (BUSWIDTH - 1))
	 while (len--) *buf++ = read8 (from++);

   return (0);
}

/*
 * Write one qword ``x'' to flash memory at offset ``offset''. ``offset''
 * must be 32 bits, i.e. it must be on a dword boundary.
 *
 * Returns 1 if successful, 0 otherwise.
 */
static inline int write_qword (__u32 offset,__u64 x)
{
   __u64 status, command, xx;

#ifdef APOLLO_DEBUG
   printk (KERN_DEBUG "%s(): 0x%.8x <- 0x%.8llx\n",__FUNCTION__,offset,x);
#endif

   /* setup writing */
   command = PGM_SETUP;
   write64 (offset, &command);

   /* write the data */
   write64 (offset, &x);

   /* wait for the write to finish */
   do
	 {
		read64 (offset, &status);
	 }
   while ((~status & STATUS_BUSY) != 0);

   /* put the flash back into command mode */
   command = READ_ARRAY;
   write64 (offset, &command);

   /* was the write successfull? */
   read64 (offset, &xx);
   if ((status & STATUS_PGM_ERR) || (xx != x))
	 {
		printk (KERN_WARNING "%s: write error at address 0x%.8x.\n",module_name,offset);
		return (0);
	 }

   return (1);
}

static int flash_write (struct mtd_info *mtd,loff_t to,size_t len,size_t *retlen,const u_char *buf)
{
   volatile __u64 tmp;
   int i,n;

#ifdef APOLLO_DEBUG
   printk (KERN_DEBUG "%s(to = 0x%.8x, len = %d)\n",__FUNCTION__,(__u32) to,len);
#endif

   *retlen = 0;

   /* sanity checks */
   if (!len) return (0);
   if (to + len > mtd->size) return (-EINVAL);

   /* first, we pad the write with current byte values until we reach a
      qword boundary */
   if (to & (BUSWIDTH - 1))
	 {
		__u32 aligned = to & ~(BUSWIDTH - 1);
		int gap = to - aligned;

		i = n = 0;

                tmp = 0;
		while (gap) tmp = (tmp << 8) | read8(to - (gap--)), i++;
		while (len && i < BUSWIDTH) tmp = (tmp << 8) | buf[n++], i++, len--;
		while (i < BUSWIDTH) tmp = (tmp << 8) | read8(aligned + i++);

		if (!write_qword (aligned, tmp)) return (-EIO);

		to += n;
		buf += n;
		*retlen += n;
	 }

   /* now we write dwords until we reach a non-dword boundary */
   while (len >= BUSWIDTH)
	 {
		if (!write_qword (to,*((__u64 *) buf))) return (-EIO);

		to += BUSWIDTH;
		buf += BUSWIDTH;
		*retlen += BUSWIDTH;
		len -= BUSWIDTH;
	 }

   /* top up the last unaligned bytes, padded with 0xFF.... */
   if (len & (BUSWIDTH - 1))
	 {
		i = n = 0;

                tmp = 0;
		while (len--) tmp = (tmp << 8) | buf[n++], i++;
		while (i < BUSWIDTH) tmp = (tmp << 8) | read8(to + (i++));

		if (!write_qword (to, tmp)) return (-EIO);

		*retlen += n;
	 }

   return (0);
}

static void flash_sync(struct mtd_info *mtd)
{

   /* mtdpart.c assumes that we have a sync function...  probably a bug,
      but easy enough to fix here instead... */

}

/***************************************************************************************************/

#define NB_OF(x) (sizeof (x) / sizeof (x[0]))

int __init apollo_flash_init (void)
{
   int result;
   memset (&mtd,0,sizeof (mtd));
   printk ("MTD driver for Apollo. Written by John W. Linville <linville@lvl7.com>\n");
   printk ("%s: Probing for 28F160C3B flash on Apollo...\n",module_name);
   if (!flash_probe ())
	 {
		printk (KERN_WARNING "%s: Found no Apollo compatible flash device\n",module_name);
		return (-ENXIO);
	 }
   printk ("%s: 28F160C3B found!\n",module_name);
   mtd.name = module_name;
   mtd.type = MTD_NORFLASH;
   mtd.flags = MTD_CAP_NORFLASH;
   mtd.size = (FLASH_BLOCKSIZE_PARAM * FLASH_NUMBLOCKS_16m_PARAM
             + FLASH_BLOCKSIZE_MAIN * FLASH_NUMBLOCKS_16m_MAIN);
   mtd.erasesize = FLASH_BLOCKSIZE_MAIN;
   mtd.numeraseregions = NB_OF (erase_regions);
   mtd.eraseregions = erase_regions;
   mtd.module = THIS_MODULE;
   mtd.erase = flash_erase;
   mtd.read = flash_read;
   mtd.write = flash_write;
   mtd.sync = flash_sync;

#ifdef APOLLO_DEBUG
   printk (KERN_DEBUG
		   "mtd.name = %s\n"
		   "mtd.size = 0x%.8x (%uM)\n"
		   "mtd.erasesize = 0x%.8x (%uK)\n"
		   "mtd.numeraseregions = %d\n",
		   mtd.name,
		   mtd.size,mtd.size / (1024*1024),
		   mtd.erasesize,mtd.erasesize / 1024,
		   mtd.numeraseregions);

   if (mtd.numeraseregions)
	 for (result = 0; result < mtd.numeraseregions; result++)
	   printk (KERN_DEBUG
			   "\n\n"
			   "mtd.eraseregions[%d].offset = 0x%.8x\n"
			   "mtd.eraseregions[%d].erasesize = 0x%.8x (%uK)\n"
			   "mtd.eraseregions[%d].numblocks = %d\n",
			   result,mtd.eraseregions[result].offset,
			   result,mtd.eraseregions[result].erasesize,mtd.eraseregions[result].erasesize / 1024,
			   result,mtd.eraseregions[result].numblocks);

#ifdef HAVE_PARTITIONS
   printk ("\npartitions = %d\n",NB_OF (apollo_partitions));

   for (result = 0; result < NB_OF (apollo_partitions); result++)
	 printk (KERN_DEBUG
			 "\n\n"
			 "apollo_partitions[%d].name = %s\n"
			 "apollo_partitions[%d].offset = 0x%.8x\n"
			 "apollo_partitions[%d].size = 0x%.8x (%uK)\n",
			 result,apollo_partitions[result].name,
			 result,apollo_partitions[result].offset,
			 result,apollo_partitions[result].size,apollo_partitions[result].size / 1024);
#endif
#endif

#ifndef HAVE_PARTITIONS
   result = add_mtd_device (&mtd);
#else
   result = add_mtd_partitions (&mtd,apollo_partitions,NB_OF (apollo_partitions));
#endif

   return (result);
}

void __exit apollo_flash_exit (void)
{
#ifndef HAVE_PARTITIONS
   del_mtd_device (&mtd);
#else
   del_mtd_partitions (&mtd);
#endif
}

module_init (apollo_flash_init);
module_exit (apollo_flash_exit);

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
MODULE_AUTHOR("John W. Linville <linville@lvl7.com>");
MODULE_DESCRIPTION("MTD driver for Intel 28F160C3B on Apollo board");
