/*
 * MTD driver for the 29LV160BB and 29LV040 Flash Memory on Broadcom BMW.
 *
 * $Id: bmw.c,v 1.1 2003/03/11 22:31:10 linville Exp $
 *
 * Author: John W. Linville <linville@lvl7.com>
 *
 * Copyright (c) 2003, LVL7 Systems, Inc.
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * References:
 *
 *    [1] Am29LV160B 16 Megabit CMOS 3.0 Volt-only Boot Sector Flash Memory
 *         -- AMD Datasheet
 *           - Publication Number: 21358
 *           - February 1999
 *
 *    [2] Am29LV040 4 Megabit CMOS 5.0 Volt-only Sector Erase Flash Memory
 *         -- AMD Datasheet
 *           - Publication Number: 21354
 *           - November 2000
 *
 *    [3] MTD internal API documentation
 *           - http://www.linux-mtd.infradead.org/tech/
 *
 * Limitations:
 *
 *    The Flash layout on this board is so bizarre, it is unlikely
 *    that this driver will be useful for any other platform.
 *
 * Notes:
 *
 *    This card has a jumper which selects between an Am29LV160BB in
 *    8-bit mode or an Am29LV040.  The same chip select is used for
 *    whichever chip is selected, and it points to the 2-meg region
 *    from 0xffe00000-0xffffffff.  When the Am29LV040 is selected, the
 *    512K address range of the device is mirrored at 0xffe00000,
 *    0xffe80000, and 0xfff80000 as well as its useful address of 
 *    0xfff00000.
 *
 *    The Am29LV040 will be ignored. (?)
 *
 */

/* debugging */
#undef BMW_DEBUG

/* partition support */
#define HAVE_PARTITIONS

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/mtd/mtd.h>
#ifdef HAVE_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

#ifndef CONFIG_BMW
#error "This is for the BMW ONLY!"
#endif

static char module_name[] = "bmw";

#define NB_OF(x) (sizeof (x) / sizeof (x[0]))

#define AM29LV160BB_ERASE_TIMEOUT	(15 * HZ)
#define AM29LV160BB_PROG_TIMEOUT	(HZ / 100)

#define AM29LV040_ERASE_TIMEOUT		(8 * HZ)
#define AM29LV040_PROG_TIMEOUT		(HZ / 100)

/*
 * These values are specific to 29LV160B flash memory.
 */
#define FLASH_BLOCKSIZE_16m_16k		16384
#define FLASH_NUMBLOCKS_16m_16k		1

#define FLASH_BLOCKSIZE_16m_8k		8192
#define FLASH_NUMBLOCKS_16m_8k		2

#define FLASH_BLOCKSIZE_16m_32k		32768
#define FLASH_NUMBLOCKS_16m_32k		1

#define FLASH_BLOCKSIZE_16m_MAIN	65536
#define FLASH_NUMBLOCKS_16m_MAIN	31
#define FLASH_NUMBLOCKS_16m_MAIN_B0_P1	15
#define FLASH_NUMBLOCKS_16m_MAIN_B0_P2_A	3
#define FLASH_NUMBLOCKS_16m_MAIN_B0_P2_B	5
#define FLASH_NUMBLOCKS_16m_MAIN_B0_P3	8

/*
 * These values are specific to 29LV040 flash memory.
 */
#define FLASH_BLOCKSIZE_4m		65536
#define FLASH_NUMBLOCKS_4m_A		3
#define FLASH_NUMBLOCKS_4m_B		5

/*
 * These values are specific to BMW.
 */

/* general */
#define FLASH_BANK0_AM29LV160BB_OFFSET	(0xffe00000)
#define FLASH_BANK0_AM29LV040_OFFSET	(0xfff00000)

#define NUM_FLASH_REGIONS	5

/* system (kernel + ramdisk) */
#define SYSTEM_START		(0x00000000)
#define SYSTEM_LEN		(0x001d0000)

/* ppcboot */
#define PPCBOOT_START		(0x001d0000)
#define PPCBOOT_LEN		(0x00030000)

/******************************************************************************/

struct flash_info {
	__u32 mtd_offset;
	__u32 dev_offset;
	size_t length;
	void *dev_base;
	struct flash_access *fl_acc;
	int num_erase_regions;
	struct mtd_erase_region_info *erase_regions;
	int erase_timeout;
	int prog_timeout;
};

struct flash_access {
	int (*erase_block)(struct flash_info *, __u32);
	void (*read)(struct flash_info *, char **, loff_t *, size_t *);
	int (*write)(struct flash_info *, char **, loff_t *, size_t *);
};

/******************************************************************************/

static int bank0_am29lv160bb_erase(struct flash_info *region, __u32 offset);
static int bank0_am29lv040_erase(struct flash_info *region, __u32 offset);

static void bank0_read(struct flash_info *region, char **to,
                       loff_t *from, size_t *len);

static int bank0_am29lv160bb_write(struct flash_info *region, char **from,
                                    loff_t *to, size_t *len);
static int bank0_am29lv040_write(struct flash_info *region, char **from,
                                 loff_t *to, size_t *len);

/******************************************************************************/

static struct flash_info fl_info[NUM_FLASH_REGIONS];

static struct flash_access flacc_bank0_am29lv160bb = {
	bank0_am29lv160bb_erase,
	bank0_read,
	bank0_am29lv160bb_write
};

static struct flash_access flacc_bank0_am29lv040 = {
	bank0_am29lv040_erase,
	bank0_read,
	bank0_am29lv040_write
};

/******************************************************************************/

static struct mtd_info mtd;

/* Offsets will be adjusted at runtime... */

static struct mtd_erase_region_info bank0_p0_erase_regions[] =
{
	/* boot blocks */
	{
		offset:    0x00000000,
		erasesize: FLASH_BLOCKSIZE_16m_16k,
		numblocks: FLASH_NUMBLOCKS_16m_16k
	},
	{
		offset:    0x00000000,
		erasesize: FLASH_BLOCKSIZE_16m_8k,
		numblocks: FLASH_NUMBLOCKS_16m_8k
	},
	{
		offset:    0x00000000,
		erasesize: FLASH_BLOCKSIZE_16m_32k,
		numblocks: FLASH_NUMBLOCKS_16m_32k
	}
};

static struct mtd_erase_region_info bank0_p1_erase_regions[] =
{
	/* main blocks */
	{
		offset:    0x00000000,
		erasesize: FLASH_BLOCKSIZE_16m_MAIN,
		numblocks: FLASH_NUMBLOCKS_16m_MAIN_B0_P1
	}
};

static struct mtd_erase_region_info bank0_p2_am29lv160bb_erase_regions_a[] =
{
        /* main blocks */
        {
                offset:    0x00000000,
                erasesize: FLASH_BLOCKSIZE_16m_MAIN,
                numblocks: FLASH_NUMBLOCKS_16m_MAIN_B0_P2_A
        }
};

static struct mtd_erase_region_info bank0_p2_am29lv160bb_erase_regions_b[] =
{
	/* main blocks */
	{
		offset:    0x00000000,
		erasesize: FLASH_BLOCKSIZE_16m_MAIN,
		numblocks: FLASH_NUMBLOCKS_16m_MAIN_B0_P2_B
	}
};

static struct mtd_erase_region_info bank0_p2_am29lv040_erase_regions_a[] =
{
	/* main blocks */
	{
		offset:    0x00000000,
		erasesize: FLASH_BLOCKSIZE_4m,
		numblocks: FLASH_NUMBLOCKS_4m_A
	}
};

static struct mtd_erase_region_info bank0_p2_am29lvf040_erase_regions_b[] =
{
	/* main blocks */
	{
		offset:    0x00000000,
		erasesize: FLASH_BLOCKSIZE_4m,
		numblocks: FLASH_NUMBLOCKS_4m_B
	}
};

static struct mtd_erase_region_info bank0_p3_erase_regions[] =
{
	/* main blocks */
	{
		offset:    0x00000000,
		erasesize: FLASH_BLOCKSIZE_16m_MAIN,
		numblocks: FLASH_NUMBLOCKS_16m_MAIN_B0_P3
	}
};

/* Here we depend on the fact that the Am29LV040 has the same erase
   characteristics as the region of the Am29LV160BB that it overlays... */

static struct mtd_erase_region_info erase_regions[
	  NB_OF(bank0_p0_erase_regions)
	+ NB_OF(bank0_p1_erase_regions)
	+ NB_OF(bank0_p2_am29lv160bb_erase_regions_a)
	+ NB_OF(bank0_p2_am29lv160bb_erase_regions_b)
	+ NB_OF(bank0_p3_erase_regions)
];

#ifdef HAVE_PARTITIONS
static struct mtd_partition bmw_partitions[] =
{
	/* system (kernel + ramdisk) */
	{
		      name: "System (Kernel + Ramdisk)",
		    offset: SYSTEM_START,
		      size: SYSTEM_LEN,
		mask_flags: 0
	},
	/* ppcboot */
	{
		      name: "PPCBoot",
		    offset: PPCBOOT_START,
		      size: PPCBOOT_LEN,
		mask_flags: 0
	}
};
#endif

/******************************************************************************/

static __u8 read8(__u32 base, __u32 offset)
{
	volatile __u8 *data = (__u8 *)(base + offset);

#ifdef BMW_DEBUG
	printk(KERN_DEBUG "%s(): 0x%.8x -> 0x%.2x\n", __FUNCTION__,
	       base + offset, *data);
#endif
	return (*data);
}

static void write8(__u32 base, __u32 offset, __u8 data)
{
#ifdef BMW_DEBUG
	printk(KERN_DEBUG "%s(): 0x%.8x -> 0x%.2x\n", __FUNCTION__,
	       base + offset, data);
#endif
	*((__u8 *)(base + offset)) = data;
}

/******************************************************************************/

static int flash_probe_bank0_am29lv160bb(void)
{
	__u8 command;

	/* Enter autoselect mode... */

	command = 0xaa;
	write8(FLASH_BANK0_AM29LV160BB_OFFSET, 0x0aaa, command);

	command = 0x55;
	write8(FLASH_BANK0_AM29LV160BB_OFFSET, 0x0555, command);

	command = 0x90;
	write8(FLASH_BANK0_AM29LV160BB_OFFSET, 0x0aaa, command);

	/* Check manufacturer... */
	command = read8(FLASH_BANK0_AM29LV160BB_OFFSET, 0x0000);

	if (command != 0x01) return (0);

	/* Check device... */
	command = read8(FLASH_BANK0_AM29LV160BB_OFFSET, 0x0002);

	if (command != 0x49) return (0);

	/* Issue reset... */

	command = 0xf0;
	write8(FLASH_BANK0_AM29LV160BB_OFFSET, 0x0000, command);

	return (1);
}

static int flash_probe_bank0_am29lv040(void)
{
	__u8 command;

	/* Enter autoselect mode... */

	command = 0xaa;
	write8(FLASH_BANK0_AM29LV040_OFFSET, 0x5555, command);

	command = 0x55;
	write8(FLASH_BANK0_AM29LV040_OFFSET, 0x2aaa, command);

	command = 0x90;
	write8(FLASH_BANK0_AM29LV040_OFFSET, 0x5555, command);

	/* Check manufacturer... */
	command = read8(FLASH_BANK0_AM29LV040_OFFSET, 0x0000);

	if (command != 0x01) return (0);

	/* Check device... */
	command = read8(FLASH_BANK0_AM29LV040_OFFSET, 0x0001);

	if (command != 0x4f) return (0);

	/* Issue reset... */

	command = 0xf0;
	write8(FLASH_BANK0_AM29LV040_OFFSET, 0x0000, command);

	return (1);
}

/*
 * Probe for flash memory on a BMW board.
 *
 * Returns 1 if we found flash memory on BMW, 0 otherwise.
 */
static int flash_probe (void)
{
	int i;
	int dev_offset = 0;
	int mtd_offset = 0;
	int num_erase_regions = 0;
	int length;

#define FLINFO_INC_OFFSETS(index) \
	length = 0; \
	for (i = 0; i < fl_info[index].num_erase_regions; i++) { \
		fl_info[index].erase_regions[i].offset = mtd_offset; \
		mtd_offset += fl_info[index].erase_regions[i].erasesize \
		            * fl_info[index].erase_regions[i].numblocks; \
		dev_offset += fl_info[index].erase_regions[i].erasesize \
		            * fl_info[index].erase_regions[i].numblocks; \
		length += fl_info[index].erase_regions[i].erasesize \
		        * fl_info[index].erase_regions[i].numblocks; \
	} \
	memcpy(&erase_regions[num_erase_regions], \
	       fl_info[index].erase_regions, \
	       fl_info[index].num_erase_regions \
	     * sizeof(erase_regions[0])); \
	num_erase_regions += fl_info[index].num_erase_regions; \
	fl_info[index].length = length;

	memset(fl_info, 0, sizeof(fl_info));
	memset(erase_regions, 0, sizeof(erase_regions));

	/* Probe for presence of bank0 (Am29LV160BB)... */

	if (flash_probe_bank0_am29lv160bb()) {

		/* Construct bank0 (Am29LV160BB) portion of fl_info... */

		fl_info[0].mtd_offset = mtd_offset;
		fl_info[0].dev_offset = dev_offset;
		fl_info[0].dev_base = (void *)FLASH_BANK0_AM29LV160BB_OFFSET;
		fl_info[0].fl_acc = &flacc_bank0_am29lv160bb;
		fl_info[0].num_erase_regions = NB_OF(bank0_p0_erase_regions);
		fl_info[0].erase_regions = bank0_p0_erase_regions;
		fl_info[0].erase_timeout = AM29LV160BB_ERASE_TIMEOUT;
		fl_info[0].prog_timeout = AM29LV160BB_PROG_TIMEOUT;

		FLINFO_INC_OFFSETS(0);

		fl_info[1].mtd_offset = mtd_offset;
		fl_info[1].dev_offset = dev_offset;
		fl_info[1].dev_base = (void *)FLASH_BANK0_AM29LV160BB_OFFSET;
		fl_info[1].fl_acc = &flacc_bank0_am29lv160bb;
		fl_info[1].num_erase_regions = NB_OF(bank0_p1_erase_regions);
		fl_info[1].erase_regions = bank0_p1_erase_regions;
		fl_info[1].erase_timeout = AM29LV160BB_ERASE_TIMEOUT;
		fl_info[1].prog_timeout = AM29LV160BB_PROG_TIMEOUT;

		FLINFO_INC_OFFSETS(1);

		dev_offset += 0x30000;

		fl_info[2].mtd_offset = mtd_offset;
		fl_info[2].dev_offset = dev_offset;
		fl_info[2].dev_base = (void *)FLASH_BANK0_AM29LV160BB_OFFSET;
		fl_info[2].fl_acc = &flacc_bank0_am29lv160bb;
		fl_info[2].num_erase_regions
		   = NB_OF(bank0_p2_am29lv160bb_erase_regions_b);
		fl_info[2].erase_regions = bank0_p2_am29lv160bb_erase_regions_b;
		fl_info[2].erase_timeout = AM29LV160BB_ERASE_TIMEOUT;
		fl_info[2].prog_timeout = AM29LV160BB_PROG_TIMEOUT;

		FLINFO_INC_OFFSETS(2);

		fl_info[3].mtd_offset = mtd_offset;
		fl_info[3].dev_offset = dev_offset;
		fl_info[3].dev_base = (void *)FLASH_BANK0_AM29LV160BB_OFFSET;
		fl_info[3].fl_acc = &flacc_bank0_am29lv160bb;
		fl_info[3].num_erase_regions = NB_OF(bank0_p3_erase_regions);
		fl_info[3].erase_regions = bank0_p3_erase_regions;
		fl_info[3].erase_timeout = AM29LV160BB_ERASE_TIMEOUT;
		fl_info[3].prog_timeout = AM29LV160BB_PROG_TIMEOUT;

		FLINFO_INC_OFFSETS(3);

		dev_offset = 0x100000;

		fl_info[4].mtd_offset = mtd_offset;
		fl_info[4].dev_offset = dev_offset;
		fl_info[4].dev_base = (void *)FLASH_BANK0_AM29LV160BB_OFFSET;
		fl_info[4].fl_acc = &flacc_bank0_am29lv160bb;
		fl_info[4].num_erase_regions
		   = NB_OF(bank0_p2_am29lv160bb_erase_regions_a);
		fl_info[4].erase_regions = bank0_p2_am29lv160bb_erase_regions_a;
		fl_info[4].erase_timeout = AM29LV160BB_ERASE_TIMEOUT;
		fl_info[4].prog_timeout = AM29LV160BB_PROG_TIMEOUT;

		FLINFO_INC_OFFSETS(4); /* Necessary to set offset value in struct... */

	} else { /* Could check for Am29LV040... (?) */

		return (0);

	}

	return (1);

#undef FLINFO_INC_OFFSETS

}

/******************************************************************************/

static int bank0_am29lv160bb_erase(struct flash_info *region, __u32 offset)
{
	unsigned long start;
	__u32 dev_offset;
	__u8 command;
	void *base;

	/* Perform sector erase... */

	dev_offset = offset - region->mtd_offset
	           + region->dev_offset;

	base = region->dev_base;

	command = 0xaa;
	write8((__u32)base, 0x0aaa, command);

	command = 0x55;
	write8((__u32)base, 0x0555, command);

	command = 0x80;
	write8((__u32)base, 0x0aaa, command);

	command = 0xaa;
	write8((__u32)base, 0x0aaa, command);

	command = 0x55;
	write8((__u32)base, 0x0555, command);

	command = 0x30;
	write8((__u32)base, dev_offset, command);

	start = jiffies + 1;

	do { /* Poll for erase complete... */

		if (jiffies > start + region->erase_timeout) return (0);

		if (current->need_resched) schedule();

		command = read8((__u32)base, dev_offset);

	} while (command != 0xff);

	return (1); /* indicate success */
}

static int bank0_am29lv040_erase(struct flash_info *region, __u32 offset)
{
	unsigned long start;
	__u32 dev_offset;
	__u8 command;
	void *base;

	/* Perform sector erase... */

	dev_offset = offset - region->mtd_offset
	           + region->dev_offset;

	base = region->dev_base;

	command = 0xaa;
	write8((__u32)base, 0x5555, command);

	command = 0x55;
	write8((__u32)base, 0x2aaa, command);

	command = 0x80;
	write8((__u32)base, 0x5555, command);

	command = 0xaa;
	write8((__u32)base, 0x5555, command);

	command = 0x55;
	write8((__u32)base, 0x2aaa, command);

	command = 0x30;
	write8((__u32)base, dev_offset, command);

	start = jiffies + 1;

	do { /* Poll for erase complete... */

		if (jiffies > start + region->erase_timeout) return (0);

		if (current->need_resched) schedule();

		command = read8((__u32)base, dev_offset);

	} while (command != 0xff);

	return (1); /* indicate success */
}

/*
 * Erase one block of flash memory at offset ``offset'' which is any
 * address within the block which should be erased.
 *
 * Returns 1 if successful, 0 otherwise.
 */
static inline int erase_block(__u32 offset)
{
	int i;

	for (i = 0; i < (NB_OF(fl_info) - 1); i++) {
		if (offset < fl_info[i + 1].mtd_offset) break;
	}
	if (i > (NB_OF(fl_info) - 1)) {
		return (0);
	} else if (i == (NB_OF(fl_info) - 1)) {
		if ((offset - fl_info[i].mtd_offset)
		  > fl_info[i].length) return (0);
	}

	return (fl_info[i].fl_acc->erase_block(&fl_info[i], offset));
}

static int flash_erase(struct mtd_info *mtd, struct erase_info *instr)
{
	__u32 addr, len;
	int i, first;

#ifdef BMW_DEBUG
	printk(KERN_DEBUG "%s(addr = 0x%.8x, len = %d)\n", __FUNCTION__,
	       instr->addr, instr->len);
#endif
 
	/* sanity checks */
	if ((instr->addr + instr->len) > mtd->size) return (-EINVAL);

	/*
	 * check that both start and end of the requested erase are
	 * aligned with the erasesize at the appropriate addresses.
	 *
	 * skip all erase regions which are ended before the start of
	 * the requested erase. Actually, to save on the calculations,
	 * we skip to the first erase region which starts after the
	 * start of the requested erase, and then go back one.
	 */
	for (i = 0;
	     ((i < mtd->numeraseregions)
           && (instr->addr >= mtd->eraseregions[i].offset));
	     i++) ;
	i--;

	/*
	 * ok, now i is pointing at the erase region in which this
	 * erase request starts. Check the start of the requested
	 * erase range is aligned with the erase size which is in
	 * effect here.
	 */
	if (instr->addr & (mtd->eraseregions[i].erasesize - 1))
		return (-EINVAL);

	/* Remember the erase region we start on */
	first = i;

	/*
	 * next, check that the end of the requested erase is aligned
	 * with the erase region at that address.
	 *
	 * as before, drop back one to point at the region in which
	 * the address actually falls
	 */
	for (;
	     ((i < mtd->numeraseregions)
	   && ((instr->addr + instr->len) >= mtd->eraseregions[i].offset));
	     i++) ;
	i--;

	/* is the end aligned on a block boundary? */
	if ((instr->addr + instr->len) & (mtd->eraseregions[i].erasesize - 1))
		return (-EINVAL);

	addr = instr->addr;
	len = instr->len;

	i = first;

	/* now erase those blocks */
	while (len) {

		if (!erase_block (addr)) {
			instr->state = MTD_ERASE_FAILED;
			return (-EIO);
		}

		addr += mtd->eraseregions[i].erasesize;
		len -= mtd->eraseregions[i].erasesize;

		if (addr
		 == mtd->eraseregions[i].offset
		  + (mtd->eraseregions[i].erasesize
		   * mtd->eraseregions[i].numblocks))
			i++;

	}

	instr->state = MTD_ERASE_DONE;
	if (instr->callback) instr->callback (instr);

	return (0);
}

/******************************************************************************/

static void bank0_read(struct flash_info *region, char **to,
                       loff_t *from, size_t *len)
{
	__u32 offset;
	__u32 length;

	offset = *from - region->mtd_offset;

	length = ((region->length - offset) >= *len)
	       ? *len : (region->length - offset);

	/* memcpy() might not work, so just do a loop (for now)... */
	while ((length--) > 0) {

		*((*to)++) = read8((__u32)region->dev_base,
		                   ((*from - region->mtd_offset)
		                  + region->dev_offset));
		(*from)++;
		(*len)--;

	}
}

static int flash_read(struct mtd_info *mtd, loff_t from, size_t len,
                      size_t *retlen, u_char *buf)
{
	int i;

#ifdef BMW_DEBUG
	printk(KERN_DEBUG "%s(from = 0x%.8x, len = %d)\n", __FUNCTION__,
               (__u32)from, len);
#endif

	/* sanity checks */
	if (!len) return (0);
	if (from + len > mtd->size) return (-EINVAL);

	/* we always read len bytes */
	*retlen = len;

	/* Determine first region accessed... */
	for (i = 0; i < (NB_OF(fl_info) - 1); i++) {

		if (from < fl_info[i + 1].mtd_offset) break;

	}
	if (i >= NB_OF(fl_info)) return (1);

	do {

		/* Call access method for region... */
		fl_info[i].fl_acc->read(&fl_info[i], (char **)&buf,
		                        &from, &len);

		/* Next region... */
		i++;

	} while ((len > 0) && (i < NB_OF(fl_info)));

	if (i >= NB_OF(fl_info)) return (1);

	return (0);
}

/******************************************************************************/

static int bank0_am29lv160bb_write(struct flash_info *region, char **from,
                                   loff_t *to, size_t *len)
{
	unsigned long start;
	__u32 offset;
	__u32 length;
	__u8 command;

	offset = (__u32)*from - region->mtd_offset;

	length = ((region->length - offset) >= *len)
	       ? *len : (region->length - offset);

	while ((length--) > 0) {

		command = 0xaa;
		write8((__u32)region->dev_base, 0x0aaa, command);

		command = 0x55;
		write8((__u32)region->dev_base, 0x0555, command);

		command = 0xa0;
		write8((__u32)region->dev_base, 0x0aaa, command);

		command = *((*from)++);
		write8((__u32)region->dev_base,
		       ((*to - region->mtd_offset) + region->dev_offset),
		       command);

		start = jiffies + 1;

		do {

			if (jiffies > start + region->prog_timeout) return (0);

			if (current->need_resched) schedule();

		} while (read8((__u32)region->dev_base,
		               ((*to - region->mtd_offset)
		              + region->dev_offset)) != command);

		(*to)++;
		(*len)--;

	}

	return (1); /* indicate success */
}

static int bank0_am29lv040_write(struct flash_info *region, char **from,
                                loff_t *to, size_t *len)
{
	unsigned long start;
	__u32 offset;
	__u32 length;
	__u8 command;

	offset = (__u32)*from - region->mtd_offset;

	length = ((region->length - offset) >= *len)
	       ? *len : (region->length - offset);

	while ((length--) > 0) {

		command = 0xaa;
		write8((__u32)region->dev_base, 0x5555, command);

		command = 0x55;
		write8((__u32)region->dev_base, 0x2aaa, command);

		command = 0xa0;
		write8((__u32)region->dev_base, 0x5555, command);

		command = *((*from)++);
		write8((__u32)region->dev_base,
		       ((*to - region->mtd_offset) + region->dev_offset),
		       command);

		start = jiffies + 1;

		do {

			if (jiffies > start + region->prog_timeout) return (0);

			if (current->need_resched) schedule();

		} while (read8((__u32)region->dev_base,
		               ((*to - region->mtd_offset)
		              + region->dev_offset)) != command);

		(*to)++;
		(*len)--;

	}

	return (1); /* indicate success */
}

static int flash_write(struct mtd_info *mtd, loff_t to, size_t len,
                       size_t *retlen, const u_char *buf)
{
	int i;

#ifdef BMW_DEBUG
	printk(KERN_DEBUG "%s(to = 0x%.8x, len = %d)\n", __FUNCTION__,
	       (__u32)to, len);
#endif

	*retlen = len;

	/* sanity checks */
	if (!len) return (0);
	if (to + len > mtd->size) return (-EINVAL);

	/* Determine first region accessed... */
	for (i = 0; i < (NB_OF(fl_info) - 1); i++) {

		if (to < fl_info[i + 1].mtd_offset) break;

	}
	if (i >= NB_OF(fl_info)) return (1);

	do {

		/* Call access method for region... */
		if (fl_info[i].fl_acc->write(&fl_info[i], (char **)&buf,
		                             &to, &len) != 0) break;

		/* Next region... */
		i++;

	} while ((len > 0) && (i < NB_OF(fl_info)));

	/* In case we didn't complete writing... */
	*retlen -= len;

	if ((len < 0) || (i >= NB_OF(fl_info))) return (-EIO);

	return (0);
}

static void flash_sync(struct mtd_info *mtd)
{

	/* mtdpart.c assumes that we have a sync function...probably a bug,
	   but easy enough to fix here instead... */

}

/******************************************************************************/

int __init bmw_flash_init(void)
{
	int i;
	int result;

	memset(&mtd, 0, sizeof (mtd));

	printk("MTD driver for BMW. Written by John W. Linville "
	        "<linville@lvl7.com>\n");

	printk("%s: Probing for BMW flash configuration...\n", module_name);
	if (!flash_probe ()) {
		printk(KERN_WARNING "%s: Found no BMW compatible flash "
		       "configuration...\n", module_name);
		return (-ENXIO);
	}
	printk ("%s: BMW flash found!\n", module_name);

	mtd.name = module_name;
	mtd.type = MTD_NORFLASH;
	mtd.flags = MTD_CAP_NORFLASH;

	for (i = 0, mtd.size = 0; i < NB_OF(fl_info); i++) {

		mtd.size += fl_info[i].length;

	}

	for (i = 0, mtd.erasesize = 0; i < NB_OF(erase_regions); i++) {

		if (erase_regions[i].erasesize > mtd.erasesize) {

			mtd.erasesize = erase_regions[i].erasesize;

		}

	}

	mtd.numeraseregions = NB_OF(erase_regions);
	mtd.eraseregions = erase_regions;
	mtd.module = THIS_MODULE;
	mtd.erase = flash_erase;
	mtd.read = flash_read;
	mtd.write = flash_write;
	mtd.sync = flash_sync;

#ifdef BMW_DEBUG
	printk(KERN_DEBUG
	       "mtd.name = %s\n"
	       "mtd.size = 0x%.8x (%uM)\n"
	       "mtd.erasesize = 0x%.8x (%uK)\n"
	       "mtd.numeraseregions = %d\n",
	       mtd.name,
	       mtd.size,mtd.size / (1024*1024),
	       mtd.erasesize,mtd.erasesize / 1024,
	       mtd.numeraseregions);

	if (mtd.numeraseregions) {

		for (result = 0; result < mtd.numeraseregions; result++) {

			printk(KERN_DEBUG
			       "\n\nmtd.eraseregions[%d].offset = 0x%.8x\n"
			       "mtd.eraseregions[%d].erasesize = 0x%.8x (%uK)"
			       "\nmtd.eraseregions[%d].numblocks = %d\n",
			       result,
			       mtd.eraseregions[result].offset,
			       result,
			       mtd.eraseregions[result].erasesize,
			       mtd.eraseregions[result].erasesize / 1024,
			       result,
			       mtd.eraseregions[result].numblocks);

		}

	}

#ifdef HAVE_PARTITIONS
	printk ("\npartitions = %d\n",NB_OF (bmw_partitions));

	for (result = 0; result < NB_OF (bmw_partitions); result++) {

		printk(KERN_DEBUG
		       "\n\nbmw_partitions[%d].name = %s\n"
		       "bmw_partitions[%d].offset = 0x%.8x\n"
		       "bmw_partitions[%d].size = 0x%.8x (%uK)\n",
		       result,
		       bmw_partitions[result].name,
		       result,
		       bmw_partitions[result].offset,
		       result,
		       bmw_partitions[result].size,
		       bmw_partitions[result].size / 1024);

	}
#endif
#endif

#ifndef HAVE_PARTITIONS
	result = add_mtd_device(&mtd);
#else
	result = add_mtd_partitions(&mtd, bmw_partitions,
	                            NB_OF(bmw_partitions));
#endif

	return (result);
}

void __exit bmw_flash_exit(void)
{
#ifndef HAVE_PARTITIONS
	del_mtd_device (&mtd);
#else
	del_mtd_partitions (&mtd);
#endif
}

module_init(bmw_flash_init);
module_exit(bmw_flash_exit);

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
MODULE_AUTHOR("John W. Linville <linville@lvl7.com>");
MODULE_DESCRIPTION("MTD driver for Am29LV160BB and Am29LV040 on BMW board");
