/*
 * $Id: psg.c,v 1.3 2003/05/05 17:58:27 linville Exp $
 *
 * Normal mappings of chips in physical memory
 *
 * Modified for Jabil PSG partitions... -- JWL
 */

/* partition support */
#define HAVE_PARTITIONS

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/config.h>
#ifdef HAVE_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

#define WINDOW_ADDR	0xff800000
#define WINDOW_SIZE	0x00800000
#define BUSWIDTH	1

#define APP_START	0x00000000
#define APP_LEN		0x00610000

#define SYSTEM_START	0x00610000
#define SYSTEM_LEN	0x000f0000

#define U_BOOT_START	0x00700000
#define U_BOOT_LEN	0x00030000

#define SYSTEM_2_START	0x00730000
#define SYSTEM_2_LEN	0x000d0000

#ifdef HAVE_PARTITIONS
static struct mtd_partition psg_partitions[] =
{
	/* app filesystem */
	{
		      name: "Application Filesystem",
		    offset: APP_START,
		      size: APP_LEN,
		mask_flags: 0
	},
	/* system (kernel + ramdisk) */
	{
		      name: "System (Kernel + Ramdisk)",
		    offset: SYSTEM_START,
		      size: SYSTEM_LEN,
		mask_flags: 0
	},
	/* u-boot */
	{
		      name: "U-Boot",
		    offset: U_BOOT_START,
		      size: U_BOOT_LEN,
		mask_flags: 0
	},
	/* system (part 2) */
	{
		      name: "System (part 2)",
		    offset: SYSTEM_2_START,
		      size: SYSTEM_2_LEN,
		mask_flags: 0
	}
};

#define NUM_PARTITIONS	(sizeof(psg_partitions) \
                       / sizeof(struct mtd_partition))
#endif

static struct mtd_info *mymtd;

__u8 psg_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

__u16 psg_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

__u32 psg_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

void psg_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

void psg_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

void psg_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

void psg_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

void psg_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio(map->map_priv_1 + to, from, len);
}

struct map_info psg_map = {
	name: "Physically mapped flash",
	size: WINDOW_SIZE,
	buswidth: BUSWIDTH,
	read8: psg_read8,
	read16: psg_read16,
	read32: psg_read32,
	copy_from: psg_copy_from,
	write8: psg_write8,
	write16: psg_write16,
	write32: psg_write32,
	copy_to: psg_copy_to
};

int __init init_psg(void)
{
       	printk(KERN_NOTICE "psg flash device: %x at %x\n", WINDOW_SIZE, WINDOW_ADDR);
	psg_map.map_priv_1 = (unsigned long)ioremap(WINDOW_ADDR, WINDOW_SIZE);

	if (!psg_map.map_priv_1) {
		printk("Failed to ioremap\n");
		return -EIO;
	}
	mymtd = do_map_probe("cfi_probe", &psg_map);
	if (mymtd) {
		mymtd->module = THIS_MODULE;

#if defined(HAVE_PARTITIONS)
		add_mtd_partitions(mymtd, psg_partitions,
		                   NUM_PARTITIONS);
#else
		add_mtd_device(mymtd);
#endif
		return 0;
	}

	iounmap((void *)psg_map.map_priv_1);
	return -ENXIO;
}

static void __exit cleanup_psg(void)
{
	if (mymtd) {
#if defined(HAVE_PARTITIONS)
		del_mtd_device(mymtd);
#else
		del_mtd_partitions(mymtd);
#endif
		map_destroy(mymtd);
	}
	if (psg_map.map_priv_1) {
		iounmap((void *)psg_map.map_priv_1);
		psg_map.map_priv_1 = 0;
	}
}

module_init(init_psg);
module_exit(cleanup_psg);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("John W. Linville <linville@lvl7.com>");
MODULE_DESCRIPTION("MTD map driver for Jabil PSG");
