/*
 * $Id: sc_omega.c,v 1.4 2003/07/03 18:08:37 linville Exp $
 *
 * Normal mappings of chips in physical memory
 *
 * Modified for SwitchCore Omega partitions... -- JWL
 */

/* partition support */
#define HAVE_PARTITIONS

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/config.h>
#ifdef HAVE_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

#define WINDOW_ADDR	0xfe000000
#define WINDOW_SIZE	0x00800000
#define BUSWIDTH	2

/* ppcboot */
#define PPCBOOT_START		(0x00000000)
#define PPCBOOT_LEN		(0x00060000)

/* system (kernel + ramdisk) */
#define SYSTEM_START		(0x00060000)
#define SYSTEM_LEN		(0x00180000)

/* vpd */
#define VPD_START               (0x001e0000)
#define VPD_LEN                 (0x00020000)

/* appplication */
#define APP_START		(0x00200000)
#define APP_LEN			(0x00600000)

#ifdef HAVE_PARTITIONS
static struct mtd_partition sc_omega_partitions[] =
{
	/* ppcboot */
	{
		      name: "PPCBoot",
		    offset: PPCBOOT_START,
		      size: PPCBOOT_LEN,
		mask_flags: 0
	},
	/* system (kernel + ramdisk) */
	{
		      name: "System (Kernel + Ramdisk)",
		    offset: SYSTEM_START,
		      size: SYSTEM_LEN,
		mask_flags: 0
	},
	/* VPD */
	{
		      name: "FASTPATH Vital Product Data",
		    offset: VPD_START,
		      size: VPD_LEN,
		mask_flags: 0
	},
	/* app filesystem */
	{
		      name: "Application Filesystem",
		    offset: APP_START,
		      size: APP_LEN,
		mask_flags: 0
	}
};

#define NUM_PARTITIONS	(sizeof(sc_omega_partitions) \
                       / sizeof(struct mtd_partition))
#endif

static struct mtd_info *mymtd;

__u8 sc_omega_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

__u16 sc_omega_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

__u32 sc_omega_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

void sc_omega_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

void sc_omega_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

void sc_omega_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

void sc_omega_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

void sc_omega_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio(map->map_priv_1 + to, from, len);
}

struct map_info sc_omega_map = {
	name: "Physically mapped flash",
	size: WINDOW_SIZE,
	buswidth: BUSWIDTH,
	read8: sc_omega_read8,
	read16: sc_omega_read16,
	read32: sc_omega_read32,
	copy_from: sc_omega_copy_from,
	write8: sc_omega_write8,
	write16: sc_omega_write16,
	write32: sc_omega_write32,
	copy_to: sc_omega_copy_to
};

int __init init_sc_omega(void)
{
       	printk(KERN_NOTICE "sc_omega flash device: %x at %x\n", WINDOW_SIZE, WINDOW_ADDR);
	sc_omega_map.map_priv_1 = (unsigned long)ioremap(WINDOW_ADDR, WINDOW_SIZE);

	if (!sc_omega_map.map_priv_1) {
		printk("Failed to ioremap\n");
		return -EIO;
	}
	mymtd = do_map_probe("cfi_probe", &sc_omega_map);
	if (mymtd) {
		mymtd->module = THIS_MODULE;

#if defined(HAVE_PARTITIONS)
		add_mtd_partitions(mymtd, sc_omega_partitions,
		                   NUM_PARTITIONS);
#else
		add_mtd_device(mymtd);
#endif
		return 0;
	}

	iounmap((void *)sc_omega_map.map_priv_1);
	return -ENXIO;
}

static void __exit cleanup_sc_omega(void)
{
	if (mymtd) {
#if defined(HAVE_PARTITIONS)
		del_mtd_device(mymtd);
#else
		del_mtd_partitions(mymtd);
#endif
		map_destroy(mymtd);
	}
	if (sc_omega_map.map_priv_1) {
		iounmap((void *)sc_omega_map.map_priv_1);
		sc_omega_map.map_priv_1 = 0;
	}
}

module_init(init_sc_omega);
module_exit(cleanup_sc_omega);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("John W. Linville <linville@lvl7.com>");
MODULE_DESCRIPTION("MTD map driver for SwitchCore Omega");
