/*
 * (C) Copyright 2001
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <mpc8xx.h>

flash_info_t	flash_info[CFG_MAX_FLASH_BANKS]; /* info for FLASH chips	*/

#if defined(CFG_ENV_IS_IN_FLASH)
# ifndef  CFG_ENV_ADDR
#  define CFG_ENV_ADDR	(CFG_FLASH_BASE + CFG_ENV_OFFSET)
# endif
# ifndef  CFG_ENV_SIZE
#  define CFG_ENV_SIZE	CFG_ENV_SECT_SIZE
# endif
# ifndef  CFG_ENV_SECT_SIZE
#  define CFG_ENV_SECT_SIZE  CFG_ENV_SIZE
# endif
#endif


/*
* Writing to Flash using a local variable as temporary storage
* was the only way to get the compiler to write to the Flash
* in 64-bit writes via the floating point registers.
*/
#define CONVERT_2_32_TO_64(x,y)  (unsigned long long )(((unsigned long long )x) | ((unsigned long long )y << 32))
#define CONVERT_4_16_TO_64(w1,w2,w3,w4) CONVERT_2_32_TO_64((unsigned int)((unsigned int)w1|(w2<<16)),(unsigned int)((unsigned int)w3|(w4<<16)) )
#define CONVERT_16_TO_64(w) CONVERT_4_16_TO_64(w,w,w,w)
#define CONVERT_8_TO_64(w) CONVERT_16_TO_64(((w) | ((w) << 16)))

static void write_via_fpu(volatile unsigned long long *addr, unsigned long long *data)
{
        __asm__ __volatile__ ("lfd  1, 0(%0)" : : "r" (data));
        __asm__ __volatile__ ("stfd 1, 0(%0)" : : "r" (addr));
        __asm__ __volatile__ ("eieio; sync");
}

static void read_via_fpu(volatile unsigned long long *addr, unsigned long long *data)
{
        
        /*
         *use the lfd instruction to pull in 64 bits at once
         *from the flash data bus
         *use different fp reg to be safe, in case these functions
         *are ever used in a multitask env 
         */ 
        __asm__ __volatile__ ("lfd  1, 0(%0)" : : "r" (addr));
       
        /*
         *now store it to the stack so that we can read it back
         *as 2 32 bit words
         */ 
        __asm__ __volatile__ ("stfd 1, 0(%0)" : : "r" (data)); 
        __asm__ __volatile__ ("eieio" : :);
        __asm__ __volatile__ ("sync"  : :);
        return data;
}

/*-----------------------------------------------------------------------
 * Functions
 */
static ulong flash_get_size (vu_long *addr, flash_info_t *info);
static int write_word (flash_info_t *info, ulong dest, unsigned long long data);
static void sector_protect(flash_info_t *info, int on, int s_start, int s_end);

/*-----------------------------------------------------------------------
 */


unsigned long flash_init (void)
{
	unsigned long size, offset = 0;
	int i;

	/* Init: enable write,
	 * or we cannot even write flash commands
	 */

	for (i=0; i<CFG_MAX_FLASH_BANKS; ++i) {
		flash_info[i].flash_id = FLASH_UNKNOWN;

		/* Re-do sizing to get full correct info */
		size = flash_get_size((vu_long *)(CFG_FLASH_BASE+offset),&flash_info[i]);

		sector_protect(&flash_info[i], 0, 0,
		               (flash_info[i].sector_count - 1));

		offset += size;
	}

#if CFG_MONITOR_BASE >= CFG_FLASH_BASE
	/* monitor protection ON by default
	 *
	 * NOTE: This assumes that monitor is in _first_ part of _first_
	 *       flash bank.
	 */

#endif

#ifdef	CFG_ENV_IS_IN_FLASH
	/* ENV protection ON by default */
	flash_protect(FLAG_PROTECT_SET,
		      CFG_ENV_ADDR,
		      CFG_ENV_ADDR+CFG_ENV_SECT_SIZE-1,
		      &flash_info[0]);
#endif
	return (offset);
}

/*-----------------------------------------------------------------------
 */
void flash_print_info  (flash_info_t *info)
{
	int i;

	if (info->flash_id == FLASH_UNKNOWN) {
		printf ("missing or unknown FLASH type\n");
		return;
	}

	switch (info->flash_id & FLASH_VENDMASK) {
	case FLASH_MAN_INTEL:	printf ("Intel ");		break;
	default:		printf ("Unknown Vendor ");	break;
	}

	switch (info->flash_id & FLASH_TYPEMASK) {
	case FLASH_28F160C3B:	printf ("28F160C3B(2 MB)\n");
				break;
	case FLASH_28F160C3T:	printf ("28F160C3T(2 MB)\n");
				break;
	case FLASH_28F320C3B:	printf ("28F320C3B(4 MB)\n");
				break;
	case FLASH_28F320C3T:	printf ("28F320C3T(4 MB)\n");
				break;
	default:		printf ("Unknown Chip Type\n");
				break;
	}

	printf ("  Size: %ld MB in %d Sectors\n",
		info->size >> 20, info->sector_count);

	printf ("  Sector Start Addresses:");
	for (i=0; i<info->sector_count; ++i) {
		if ((i % 5) == 0)
			printf ("\n   ");
		printf (" %08lX%s",
			info->start[i],
			info->protect[i] ? " (RO)" : "     "
		);
	}
	printf ("\n");
}

/*-----------------------------------------------------------------------
 */


/*-----------------------------------------------------------------------
 */

/*
 * The following code cannot be run from FLASH!
 */

static ulong flash_get_size (vu_long *addr, flash_info_t *info)
{
	short i;
	ulong value;
        volatile unsigned long flash_bus_value;
	ulong base = (ulong)addr;
	short bottom_boot_cnt = 0, top_boot_cnt = 0;

	/* Write "Intelligent Identifier" command: read Manufacturer ID */
        /*
         *Notice that the apollo board has a 64 bit data bus to
         *the flash.  This means we need to access all 4 flash chips
         *in one write cycle to get the proper identifiers back ok
         *which means we need to use unsigned long longs here
         */
        flash_bus_value = 0x90909090; 
	*addr = flash_bus_value;

	value = addr[0];
         
	switch (value & 0x00FF0000)
	{
	case (MT_MANUFACT & 0x00FF0000):
	case (INTEL_ALT_MANU & 0x00FF0000):	/* MT or => Intel */
		info->flash_id = FLASH_MAN_INTEL;
		break;
	default:
		info->flash_id = FLASH_UNKNOWN;
		info->sector_count = 0;
		info->size = 0;
		return (0);			/* no or unknown flash	*/
	}

       
        /*
         *Again, since we have a 64 bit data 
         *bus we have 32 bit offsets [0] and [1] holding
         *the Manufacture code for all 4 chips
         *followed by the device codes for 
         *all 4 chips starting at offset [2]
         *and going through offset [3]
         *note this code also assumes that
         *each set of 2 chips (aligned) is 
         *the same, as we never check the 
         *odd numbered entries.  This is 
         *probably ok, but keep an eye on it
         */ 
        value = addr[2]; /*get the device code*/
	switch (value & 0x0000FFFF) {
	case (INTEL_ID_28F160C3B & 0x0000FFFF):
                /*
                 *This is the apollo device, which is configured
                 *as 1 Million addresses x 64 bits in 1 flash bank
                 */
		info->flash_id += FLASH_28F160C3B;
		info->sector_count = 39;
                /*
                 *our size is 1MB address * 16 bits/chip * 4 parallel chips
                 *for a total of 64 MB
                 */
		info->size = 0x00800000;
		bottom_boot_cnt = 8;
		break;				/* => 64 MB - 8MB addrs x 8bytes /addr		*/

	case (INTEL_ID_28F160C3T & 0x0000FFFF):
		info->flash_id += FLASH_28F160C3T;
		info->sector_count = 39;
		info->size = 0x00200000;
		top_boot_cnt = 8;
		break;				/* => 2 MB		*/

	case (INTEL_ID_28F320C3B & 0x0000FFFF):
		info->flash_id += FLASH_28F320C3B;
		info->sector_count = 71;
		info->size = 0x00400000;
		bottom_boot_cnt = 8;
		break;				/* => 4 MB		*/

	case (INTEL_ID_28F320C3T & 0x0000FFFF):
		info->flash_id += FLASH_28F320C3T;
		info->sector_count = 71;
		info->size = 0x00400000;
		top_boot_cnt = 8;
		break;				/* => 4 MB		*/

	default:
		info->flash_id = FLASH_UNKNOWN;
		return (0);			/* => no or unknown flash */

	}

	/* set up sector start address table */
	/*Should be done through the region codes.*/

	for(i = 0; i < bottom_boot_cnt; i++) {
		//first 8 sectors are 8KB
                info->start[i] = base + (i * 0x8000);
		info->protect[i] = 0;
	}
        
        base += (bottom_boot_cnt * 0x8000);
	for (; i < (info->sector_count - top_boot_cnt); i++) {
		//Next 31 or 63 sectors
		info->start[i] = base + ((i - bottom_boot_cnt) * 0x40000);
		/* don't know how to check sector protection */
		info->protect[i] = 0;
	}
	base += ((info->sector_count - bottom_boot_cnt - top_boot_cnt)
	        * 0x40000);
	for(; i < info->sector_count; i++) {
		//last 8 sectors are 8KB
		ulong multiplier_fixup = info->sector_count - top_boot_cnt;
		info->start[i] = base + ((i - multiplier_fixup) * 0x8000);
		info->protect[i] = 0;
	}

	/*
	 * Prevent writes to uninitialized FLASH.
	 */
	if (info->flash_id != FLASH_UNKNOWN) {
		addr = (vu_long *)info->start[0];

		*addr = 0xFFFFFFFF;	/* reset bank to read array mode */
	}

	return (info->size);
}

/*-----------------------------------------------------------------------*/
#define INTEL_FLASH_STATUS_BLS	0x02
#define INTEL_FLASH_STATUS_PSS	0x04
#define INTEL_FLASH_STATUS_VPPS	0x08
#define INTEL_FLASH_STATUS_PS	0x10
#define INTEL_FLASH_STATUS_ES	0x20
#define INTEL_FLASH_STATUS_ESS	0x40
#define INTEL_FLASH_STATUS_WSMS	0x80

static int flash_decode_status_bits(unsigned long long status)
{
	int err = 0;
	
	if(!(status & CONVERT_16_TO_64(INTEL_FLASH_STATUS_WSMS))) {
		printf("Busy\n");
		err = -1;
	}
  
	if(status & CONVERT_16_TO_64(INTEL_FLASH_STATUS_ESS)) {
		printf("Erase suspended\n");
		err = -1;
	}
 
	if(status & CONVERT_16_TO_64(INTEL_FLASH_STATUS_ES)) {
		printf("Error in block erase\n");
		err = -1;
	}

	if(status &  CONVERT_16_TO_64(INTEL_FLASH_STATUS_PS)) {
		printf("Error in programming\n");
		err = -1;
	}
        
	if(status & CONVERT_16_TO_64(INTEL_FLASH_STATUS_VPPS)) {
		printf("Vpp low, operation aborted\n");
		err = -1;
	}

	if(status & CONVERT_16_TO_64(INTEL_FLASH_STATUS_PSS)) {
		printf("Program is suspended\n");
		err = -1;
	}

	if(status & CONVERT_16_TO_64(INTEL_FLASH_STATUS_BLS)) {
		printf("Attempting to program/erase a locked sector\n");
		err = -1;
	}

	if((status & CONVERT_16_TO_64(INTEL_FLASH_STATUS_PS)) &&
	   (status & CONVERT_16_TO_64(INTEL_FLASH_STATUS_ES)) &&
	   (status & CONVERT_16_TO_64(INTEL_FLASH_STATUS_ESS))) {
		printf("A command sequence error\n");
		return -1;
	}

	return err;
}

/*-----------------------------------------------------------------------
 */

int	flash_erase (flash_info_t *info, int s_first, int s_last)
{
        volatile unsigned long long *addr = 0;
        unsigned long long data;

	int flag, prot, sect;
	ulong start, now, last = 0;

	if ((s_first < 0) || (s_first > s_last)) {
		if (info->flash_id == FLASH_UNKNOWN) {
			printf ("- missing\n");
		} else {
			printf ("- no sectors to erase\n");
		}
		return 1;
	}

	if ((info->flash_id & FLASH_VENDMASK) != FLASH_MAN_INTEL) {
		printf ("Can't erase unknown flash type %08lx - aborted\n",
			info->flash_id);
	return 1;
	}

	prot = 0;
	for (sect=s_first; sect<=s_last; ++sect) {
		if (info->protect[sect]) {
			prot++;
		}
	}

	if (prot) {
		printf ("- Warning: %d protected sectors will not be erased!\n",
			prot);
	} else {
		printf ("\n");
	}

	start = get_timer (0);

	/* Start erase on unprotected sectors */
	for (sect = s_first; sect<=s_last; sect++) {
		unsigned long long tmp;
                unsigned long long status_mask = CONVERT_16_TO_64(INTEL_FLASH_STATUS_WSMS);

		if (info->protect[sect] == 0) {	/* not protected */
			addr = (volatile unsigned long long *)(info->start[sect]);
                       
			/* Disable interrupts which might cause a timeout here */
			flag = disable_interrupts();

			/* Clear status register */
                        data = 0x0050005000500050;
                        write_via_fpu(addr,&data);
			/* Single Block Erase Command */
		        data = 0x0020002000200020;
                        write_via_fpu(addr,&data); 
			/* Confirm */
                        data = 0x00D000D000D000D0;
                        write_via_fpu(addr,&data);
			/* Resume Command, as per errata update */
		        write_via_fpu(addr,&data);	
			/* Read status register */
                        data = 0x0070007000700070;
		        write_via_fpu(addr,&data);	
			
                        /* re-enable interrupts if necessary */
			if (flag)
				enable_interrupts();

			read_via_fpu(addr,&tmp);
			while (!((tmp & status_mask) == status_mask)) {
				if ((now=get_timer(start)) > CFG_FLASH_ERASE_TOUT) {
					printf ("Timeout\n");
                                        data = 0x00FF00FF00FF00FF;
				        write_via_fpu(addr,&data);	/* reset bank */
					return 1;
				}
				/* show that we're waiting */
				if ((now - last) > 1000) {	/* every second */
					putc ('.');
					last = now;
				}
				udelay(80);
				read_via_fpu(addr,&tmp);
				if(flash_decode_status_bits(tmp|CONVERT_16_TO_64(INTEL_FLASH_STATUS_WSMS))==-1
						&& (tmp & CONVERT_16_TO_64(INTEL_FLASH_STATUS_WSMS))) {
					printf("Error erasing\n");
					break;
				}
			}
		}
	}
	/* Clear status register */
        data = 0x0050005000500050;
        write_via_fpu(addr,&data); 
	/* reset to read mode */
        data = 0x00FF00FF00FF00FF;
        write_via_fpu(addr,&data);
	return 0;
}



void	sector_protect (flash_info_t *info, int on, int s_first, int s_last)
{
        volatile unsigned long long *addr=0;
	vu_short status;
	int flag, sect;
        unsigned long long cmd;
        unsigned long long temp_data;

	if ((s_first < 0) || (s_first > s_last)) {
		return;
	}

	if ((info->flash_id & FLASH_VENDMASK) != FLASH_MAN_INTEL) {
		printf ("Can't erase unknown flash type %08lx - aborted\n",
			info->flash_id);
		return;
	}

	if(on == 0) {
		//Unlock the sector
                cmd = CONVERT_16_TO_64(0x00D0);
	}
	else {
		//Lock the sector
		cmd = CONVERT_16_TO_64(0x0010);
	}
	
	for (sect = s_first; sect<=s_last; sect++) {
		addr = (volatile unsigned long long *)(info->start[sect]);

		/* Disable interrupts which might cause a timeout here */
		flag = disable_interrupts();
               
                temp_data = CONVERT_16_TO_64(0x0050); 
                write_via_fpu(addr,&temp_data); 
                __asm__("        eieio;  sync");
                /* Config setup */
                temp_data = CONVERT_16_TO_64(0x0060);
                write_via_fpu(addr,&temp_data); 
                __asm__("        eieio;  sync");
                /* (Un)lock block */
                write_via_fpu(addr,&cmd);
                __asm__("        eieio;  sync");
                /* Read Config */
                temp_data = CONVERT_16_TO_64(0x0090);
                write_via_fpu(addr,&temp_data);
                __asm__("        eieio;  sync");
		
                /* Confirm */
		status = *((volatile unsigned short *)addr + 0x8);
		if ((status & 0x0001) != 0)
			printf("Flash sector 0x%x still locked!\n",
			       (unsigned int)addr);

		/* re-enable interrupts if necessary */
		if (flag)
			enable_interrupts();
	}
        /* Clear status register */
        *addr = 0x00500050;
        /* reset to read mode */
        *addr = 0x00FF00FF;
}

/*-----------------------------------------------------------------------
 * Copy memory to flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 */

int write_buff (flash_info_t *info, uchar *src, ulong addr, ulong cnt)
{
        ulong cp,wp;
        unsigned long long data;
	int i, l, rc;

	wp = (addr & ~7);	/* get lower word aligned address */

	/*
	 * handle unaligned start bytes
	 */
	if ((l = addr - wp) != 0) {
		data = 0;
		for (i=0, cp=wp; i<l; ++i, ++cp) {
			data = (data << 8) | (*(uchar *)cp);
		}
		for (; i<8 && cnt>0; ++i) {
			data = (data << 8) | *src++;
			--cnt;
			++cp;
		}
		for (; cnt==0 && i<8; ++i, ++cp) {
			data = (data << 8) | (*(uchar *)cp);
		}
		if ((rc = write_word(info, wp, data)) != 0) {
			return (rc);
		}
		wp += 8;
	}

	/*
	 * handle word aligned part
	 */
	while (cnt >= 8) {
		data = 0;
		for (i=0; i<8; ++i) {
			data = (data << 8) | *src++;
		}
		if ((rc = write_word(info, wp, data)) != 0) {
			return (rc);
		}
		wp  += 8;
		cnt -= 8;
	}

	if (cnt == 0) {
		return (0);
	}

	/*
	 * handle unaligned tail bytes
	 */
	data = 0;
	for (i=0, cp=wp; i<8 && cnt>0; ++i, ++cp) {
		data = (data << 8) | *src++;
		--cnt;
	}

	for (; i<8; ++i, ++cp) {
		data = (data << 8) | (*(uchar *)cp);
	}

	return (write_word(info, wp, data));
}

/*-----------------------------------------------------------------------
 * Write a word to Flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 */
static int write_word (flash_info_t *info, ulong dest, unsigned long long da)
{
	volatile unsigned long long *addr = (volatile unsigned long *)dest;
	ulong start;
	unsigned long long csr;
	int flag;
        unsigned long long cmd_data;
	unsigned long long * d = (unsigned long long*)&da;

	/* Check if Flash is (sufficiently) erased */
	if (*addr != 0xFFFFFFFFFFFFFFFF) {
		return (2);
	}
	/* Disable interrupts which might cause a timeout here */
	flag = disable_interrupts();


	/* Clear status register */
        cmd_data = 0x0050005000500050;
        write_via_fpu(addr,&cmd_data);
		
        /* Write Command */
        cmd_data = 0x0040004000400040;
        write_via_fpu(addr,&cmd_data);
               
        /* Write Data */
	write_via_fpu(addr,d);

        /* Read status register */
#if 0
        cmd_data = 0x0070007000700070;
        write_via_fpu(addr,&cmd_data);
#endif
        /* re-enable interrupts if necessary */
        if (flag)
           enable_interrupts();

	/* data polling for D7 */
	start = get_timer (0);
	flag  = 0;
        read_via_fpu(addr,&csr);
	while ((csr & CONVERT_16_TO_64(INTEL_FLASH_STATUS_WSMS))!=CONVERT_16_TO_64(INTEL_FLASH_STATUS_WSMS)) {
		if (get_timer(start) > CFG_FLASH_WRITE_TOUT) {
			flag = 1;
			break;
		}
                read_via_fpu(addr,&csr);
	}
	if (csr & CONVERT_16_TO_64(INTEL_FLASH_STATUS_PSS)) {
		printf ("CSR indicates write error (%0lx) at %08lx\n",
				csr, (ulong)addr);
		flag = 1;
	}

	/* Clear Status Registers Command */
               
        cmd_data = 0x0050005000500050;
        write_via_fpu(addr,&cmd_data);
	/* Reset to read array mode */

        cmd_data = 0x00FF00FF00FF00FF;
        write_via_fpu(addr,&cmd_data);

	return (flag);
}

/*-----------------------------------------------------------------------
 */
