/*
 * MOUSSE/MPC8240 Board definitions.
 * Flash Routines for MOUSSE onboard AMD29LV160BB and AM29F040 devices
 *
 * (C) Copyright 2003
 * John W. Linville, LVL7 Systems, Inc., linville@lvl7.com.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <asm/io.h>
#include "mousse.h"

#define FLASH_SECTOR_SIZE		(64*1024)

#define FLASH_BANK0_AM29LV160BB_BASE	0xffe00000
#define FLASH_BANK0_AM29F040_BASE	0xfff00000

#define FLASH_BANK0_P0_BASE		0xffe00000
#define FLASH_BANK0_P1_BASE		0xffe10000
#define FLASH_BANK0_P2_BASE		0xfff00000
#define FLASH_BANK0_P3_BASE		0xfff80000

#define FLASH_BANK0_P0_SIZE		0x00010000
#define FLASH_BANK0_P1_SIZE		0x000f0000
#define FLASH_BANK0_P2_SIZE		0x00080000
#define FLASH_BANK0_P3_SIZE		0x00080000

#define FLASH_BANK0_P0_SECTOR_COUNT	3
#define FLASH_BANK0_P1_SECTOR_COUNT	15
#define FLASH_BANK0_P2_SECTOR_COUNT	8
#define FLASH_BANK0_P3_SECTOR_COUNT	8

#define PAUSE_QUANTA			166 /* usec */

#define AM29LV160BB_ERASE_TIMEOUT	(15000000 / PAUSE_QUANTA)
#define AM29LV160BB_PROG_TIMEOUT	(10000 / PAUSE_QUANTA)

#define AM29F040_ERASE_TIMEOUT		(8000000 / PAUSE_QUANTA)
#define AM29F040_PROG_TIMEOUT		(10000 / PAUSE_QUANTA)

/*-----------------------------------------------------------------------
 */

struct flash_data {
	__u32 dev_offset;
	void *dev_base;
	int (*erase)(struct flash_data *, __u32);
	int (*write)(struct flash_data *, char *, ulong, ulong);
	int erase_timeout;
	int prog_timeout;
};

/*-----------------------------------------------------------------------
 */

static int flash_erase_am29lv160bb(struct flash_data *, __u32);
static int flash_write_am29lv160bb(struct flash_data *, char *, ulong, ulong);

static int flash_erase_am29f040(struct flash_data *, __u32);
static int flash_write_am29f040(struct flash_data *, char *, ulong, ulong);

/*-----------------------------------------------------------------------
 */

flash_info_t flash_info[CFG_MAX_FLASH_BANKS]; /* info for FLASH chips */

static struct flash_data fldata[CFG_MAX_FLASH_BANKS]; /* info for FLASH chips */

/*-----------------------------------------------------------------------
 */

static int flash_probe_am29lv160bb(void)
{
	__u8 command;

	/* Enter autoselect mode... */

	command = 0xaa;
	out_8((volatile u8 *)(FLASH_BANK0_AM29LV160BB_BASE + 0x0aaa), command);

	command = 0x55;
	out_8((volatile u8 *)(FLASH_BANK0_AM29LV160BB_BASE + 0x0555), command);

	command = 0x90;
	out_8((volatile u8 *)(FLASH_BANK0_AM29LV160BB_BASE + 0x0aaa), command);

	/* Check manufacturer... */
	command = in_8((volatile u8 *)(FLASH_BANK0_AM29LV160BB_BASE + 0x0000));

	if (command != 0x01) return (1);

	/* Check device... */
	command = in_8((volatile u8 *)(FLASH_BANK0_AM29LV160BB_BASE + 0x0002));

	if (command != 0x49) return (1);

	/* Issue reset... */

	command = 0xf0;
	out_8((volatile u8 *)(FLASH_BANK0_AM29LV160BB_BASE + 0x0000), command);

	return (0);
}

static int flash_probe_am29f040(void)
{
	__u8 command;

	/* Enter autoselect mode... */

	command = 0xaa;
	out_8((volatile u8 *)(FLASH_BANK0_AM29F040_BASE + 0x5555), command);

	command = 0x55;
	out_8((volatile u8 *)(FLASH_BANK0_AM29F040_BASE + 0x2aaa), command);

	command = 0x90;
	out_8((volatile u8 *)(FLASH_BANK0_AM29F040_BASE + 0x5555), command);

	/* Check manufacturer... */
	command = in_8((volatile u8 *)FLASH_BANK0_AM29F040_BASE + 0x0000);

	if (command != 0x01) return (1);

	/* Check device... */
	command = in_8((volatile u8 *)FLASH_BANK0_AM29F040_BASE + 0x0001);

	if (command != 0xa4) return (1);

	/* Issue reset... */

	command = 0xf0;
	out_8((volatile u8 *)(FLASH_BANK0_AM29F040_BASE + 0x0000), command);

	return (0);
}

unsigned long flash_init(void)
{
	int i;

	/* Init: no FLASHes known */
	for (i=0; i<CFG_MAX_FLASH_BANKS; i++) {
		flash_info[i].flash_id = FLASH_UNKNOWN;
	}

	/* DETECT and record flash configuration... */

	if (flash_probe_am29lv160bb() != 0) {

		return (0);

	}

	flash_info[0].size = FLASH_BANK0_P1_SIZE;
	flash_info[0].sector_count = FLASH_BANK0_P1_SECTOR_COUNT;
	flash_info[0].flash_id = (FLASH_MAN_AMD | FLASH_AM160LV);

	for (i=0; i<CFG_MAX_FLASH_SECT; i++) {
		flash_info[0].start[i] = 0;
	}

	for (i=0; i<flash_info[0].sector_count; i++) {
		flash_info[0].start[i] = FLASH_BANK0_P1_BASE
		                    + (i * FLASH_SECTOR_SIZE);
	}

	for (i=0; i<CFG_MAX_FLASH_SECT; i++) {
		flash_info[0].protect[i] = 0;
	}

	fldata[0].dev_offset = 0;
	fldata[0].dev_base = (void *)FLASH_BANK0_AM29LV160BB_BASE;
	fldata[0].erase = flash_erase_am29lv160bb;
	fldata[0].write = flash_write_am29lv160bb;
	fldata[0].erase_timeout = AM29LV160BB_ERASE_TIMEOUT;
	fldata[0].prog_timeout = AM29LV160BB_PROG_TIMEOUT;

	flash_info[1].size = FLASH_BANK0_P2_SIZE;
	flash_info[1].sector_count = FLASH_BANK0_P2_SECTOR_COUNT;
	flash_info[1].flash_id = (FLASH_MAN_AMD | FLASH_AM160LV);

	for (i=0; i<CFG_MAX_FLASH_SECT; i++) {
		flash_info[1].start[i] = 0;
	}

	for (i=0; i<flash_info[1].sector_count; i++) {
		flash_info[1].start[i] = FLASH_BANK0_P2_BASE
		                    + (i * FLASH_SECTOR_SIZE);
	}

	for (i=0; i<CFG_MAX_FLASH_SECT; i++) {
		flash_info[1].protect[i] = 0;
	}

	fldata[1].dev_offset = FLASH_BANK0_P1_SIZE;
	fldata[1].dev_base = (void *)FLASH_BANK0_AM29LV160BB_BASE;
	fldata[1].erase = flash_erase_am29lv160bb;
	fldata[1].write = flash_write_am29lv160bb;
	fldata[1].erase_timeout = AM29LV160BB_ERASE_TIMEOUT;
	fldata[1].prog_timeout = AM29LV160BB_PROG_TIMEOUT;

	flash_info[2].size = FLASH_BANK0_P3_SIZE;
	flash_info[2].sector_count = FLASH_BANK0_P3_SECTOR_COUNT;
	flash_info[2].flash_id = (FLASH_MAN_AMD | FLASH_AM160LV);

	for (i=0; i<CFG_MAX_FLASH_SECT; i++) {
		flash_info[2].start[i] = 0;
	}

	for (i=0; i<flash_info[2].sector_count; i++) {
		flash_info[2].start[i] = FLASH_BANK0_P3_BASE
		                    + (i * FLASH_SECTOR_SIZE);
	}

	for (i=0; i<CFG_MAX_FLASH_SECT; i++) {
		flash_info[2].protect[i] = 0;
	}

	fldata[2].dev_offset = FLASH_BANK0_P1_SIZE
	                     + FLASH_BANK0_P2_SIZE;
	fldata[2].dev_base = (void *)FLASH_BANK0_AM29LV160BB_BASE;
	fldata[2].erase = flash_erase_am29lv160bb;
	fldata[2].write = flash_write_am29lv160bb;
	fldata[2].erase_timeout = AM29LV160BB_ERASE_TIMEOUT;
	fldata[2].prog_timeout = AM29LV160BB_PROG_TIMEOUT;

	if (flash_probe_am29f040() == 0) {

		flash_info[1].flash_id = (FLASH_MAN_AMD | FLASH_AM040);

		fldata[1].dev_offset = 0;
		fldata[1].dev_base = (void *)FLASH_BANK0_AM29F040_BASE;
		fldata[1].erase = flash_erase_am29f040;
		fldata[1].write = flash_write_am29f040;
		fldata[1].erase_timeout = AM29F040_ERASE_TIMEOUT;
		fldata[1].prog_timeout = AM29F040_PROG_TIMEOUT;


	}

#if CFG_MONITOR_BASE >= CFG_FLASH_BASE
	/* monitor protection ON by default */
	flash_protect(FLAG_PROTECT_SET,
	              CFG_MONITOR_BASE,
	              CFG_MONITOR_BASE+CFG_MONITOR_LEN-1,
	              &flash_info[0]);
#endif

	return (FLASH_BANK0_P1_SIZE
	      + FLASH_BANK0_P2_SIZE
	      + FLASH_BANK0_P3_SIZE);
}

void
flash_print_info  (flash_info_t *info)
{
	int i;
	unsigned int chip;

	if (info->flash_id == FLASH_UNKNOWN) {
		printf ("missing or unknown FLASH type\n");
		return;
	}

	switch (info->flash_id >> 16) {
	case FLASH_MAN_AMD:
		printf ("AMD ");
		break;
	default:
		printf ("Unknown Vendor ");
		break;
	}
	chip = (unsigned int) info->flash_id & 0x0000ffff;

	switch (chip) {
	case FLASH_AM040:
		printf ("AM29F040 (4 Mbit / 512K x 8bit)\n");
		break;

	case FLASH_AM160LV:
		printf ("AM29LV160BB (16 Mbit / 2M x 8bit)\n");
		break;

	default:
		printf ("Unknown Chip Type:0x%x\n", chip);
		break;
	}

	printf("  Size: %ld bytes in %d Sectors\n",
	       info->size, info->sector_count);

	printf ("  Sector Start Addresses:");
	for (i=0; i<info->sector_count; ++i) {
		if ((i % 5) == 0)
			printf ("\n   ");
		printf(" %08lX%s",
		       info->start[0] + i*FLASH_SECTOR_SIZE,
		       info->protect[i] ? " (RO)" : "     ");
	}
	printf ("\n");
}

/*-----------------------------------------------------------------------
 */

static int flash_erase_am29lv160bb(struct flash_data *region, __u32 addr)
{
	unsigned long loops = 0;
	__u8 command;
	void *base;

	/* Perform sector erase... */

	base = region->dev_base;

	command = 0xaa;
	out_8((volatile u8 *)(base + 0x0aaa), command);

	command = 0x55;
	out_8((volatile u8 *)(base + 0x0555), command);

	command = 0x80;
	out_8((volatile u8 *)(base + 0x0aaa), command);

	command = 0xaa;
	out_8((volatile u8 *)(base + 0x0aaa), command);

	command = 0x55;
	out_8((volatile u8 *)(base + 0x0555), command);

	command = 0x30;
	out_8((volatile u8 *)addr, command);

	do { /* Poll for erase complete... */

		udelay(PAUSE_QUANTA);

		command = in_8((volatile u8 *)addr);

	} while ((command != 0xff) && (loops++ <= region->erase_timeout));

	if (loops > region->erase_timeout) return (1);

	return (0); /* indicate success */
}

static int flash_erase_am29f040(struct flash_data *region, __u32 addr)
{
	unsigned long loops = 0;
	__u8 command;
	void *base;

	/* Perform sector erase... */

	base = region->dev_base;

	command = 0xaa;
	out_8((volatile u8 *)(base + 0x5555), command);

	command = 0x55;
	out_8((volatile u8 *)(base + 0x2aaa), command);

	command = 0x80;
	out_8((volatile u8 *)(base + 0x5555), command);

	command = 0xaa;
	out_8((volatile u8 *)(base + 0x5555), command);

	command = 0x55;
	out_8((volatile u8 *)(base + 0x2aaa), command);

	command = 0x30;
	out_8((volatile u8 *)addr, command);

	do { /* Poll for erase complete... */

		udelay(PAUSE_QUANTA);

		command = in_8((volatile u8 *)addr);

	} while ((command != 0xff) && (loops++ <= region->erase_timeout));

	if (loops > region->erase_timeout) return (1);

	return (0); /* indicate success */
}

/*-----------------------------------------------------------------------
 * Erase a range of flash sectors.
 * 0 - OK
 * 1 - erase failure/timeout
 */
int flash_erase (flash_info_t *info, int s_first, int s_last)
{
	int i;
	int bank;

	/* Determine bank number... */
	bank = (((__u32)info - (__u32)&flash_info[0]) / sizeof(*info));

	for (i = s_first; i <= s_last; i++) {

		if ((*fldata[bank].erase)(&fldata[bank], info->start[i]) != 0) {

			return (1);

		}

		putc('.');

	}

	putc('\n');

	return (0);
}

/*-----------------------------------------------------------------------
 */

static int flash_write_am29lv160bb(struct flash_data *region, char *from,
                                   ulong to, ulong len)
{
	unsigned long loops = 0;
	__u8 command;

	while ((len--) > 0) {

		/* Check for sufficient erasure... */
		if ((*((char *)to) & *from) != *from) return (2);

		command = 0xaa;
		out_8((volatile u8 *)(region->dev_base + 0x0aaa), command);

		command = 0x55;
		out_8((volatile u8 *)(region->dev_base + 0x0555), command);

		command = 0xa0;
		out_8((volatile u8 *)(region->dev_base + 0x0aaa), command);

		command = *(from++);
		out_8((volatile u8 *)to, command);

		loops = 0;

		while ((in_8((volatile u8 *)to) != command)
		    && (loops++ <= region->prog_timeout)) /* empty */;

		if (loops > region->erase_timeout) return (1);

		to++;

	}

	return (0);
}

static int flash_write_am29f040(struct flash_data *region, char *from,
                                ulong to, ulong len)
{
	unsigned long loops = 0;
	__u8 command;

	while ((len--) > 0) {

		/* Check for sufficient erasure... */
		if ((*((char *)to) & *from) != *from) return (2);

		command = 0xaa;
		out_8((volatile u8 *)(region->dev_base + 0x5555), command);

		command = 0x55;
		out_8((volatile u8 *)(region->dev_base + 0x2aaa), command);

		command = 0xa0;
		out_8((volatile u8 *)(region->dev_base + 0x5555), command);

		command = *(from++);
		out_8((volatile u8 *)to, command);

		loops = 0;

		while ((in_8((volatile u8 *)to) != command)
		    && (loops++ <= region->prog_timeout)) /* empty */;

		if (loops > region->erase_timeout) return (1);

		to++;

	}

	return (0);
}

/*-----------------------------------------------------------------------
 * Copy memory to flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 */

int write_buff (flash_info_t *info, uchar *src, ulong addr, ulong cnt)
{
	int bank;

	/* Determine bank number... */
	bank = (((__u32)info - (__u32)&flash_info[0]) / sizeof(*info));

	return ((*fldata[bank].write)(&fldata[bank], src, addr, cnt));
}

/*-----------------------------------------------------------------------
 */
