/*
 * MOUSSE Board Support -- Flash bank 1 access
 *
 * (C) Copyright 2002
 * John W. Linville, linville@lvl7.com
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <command.h>

#define NB_OF(x)		(sizeof(x) / sizeof(x[0]))

#define FLASH_BANK1_BASE	0xff000000

struct sect_info {
	int num_sects;
	int sect_size;
};

#define FLASH_BANK1_POLL_DELAY		133
#define FLASH_BANK1_ERASE_TIMEOUT	(((15 * 1000000) \
                                        / FLASH_BANK1_POLL_DELAY) + 1)
#define FLASH_BANK1_WRITE_TIMEOUT	((360 / FLASH_BANK1_POLL_DELAY) + 1)

#if (CONFIG_COMMANDS & CFG_CMD_BSP)

static struct sect_info sinfo[] = {
	{  1, 0x04000},
	{  2, 0x02000},
	{  1, 0x08000},
	{ 31, 0x10000}
};

static void write_via_fpu(volatile unsigned long long *addr,
                          unsigned long long *data)
{
#if 1 /*JWL -- why doesn't this work??? */
	ulong tmp;

	tmp = (ulong)(*data >> 32);

	*((ulong *)addr) = tmp;
#else
        __asm__ __volatile__ ("lfd  1, 0(%0)" : : "r" (data));
        __asm__ __volatile__ ("stfd 1, 0(%0)" : : "r" (addr));
        __asm__ __volatile__ ("eieio; sync");
#endif
}

static void read_via_fpu(volatile unsigned long long *addr,
                         unsigned long long *data)
{
        /*
         *use the lfd instruction to pull in 64 bits at once
         *from the flash data bus
         *use different fp reg to be safe, in case these functions
         *are ever used in a multitask env
         */
        __asm__ __volatile__ ("lfd  1, 0(%0)" : : "r" (addr));

        /*
         *now store it to the stack so that we can read it back
         *as 2 32 bit words
         */
        __asm__ __volatile__ ("stfd 1, 0(%0)" : : "r" (data));
        __asm__ __volatile__ ("eieio" : :);
        __asm__ __volatile__ ("sync"  : :);
}

static int flbank1_load(cmd_tbl_t *cmdtp, int argc, char *argv[])
{
	char *s;
	int i;
	ulong checksum;
	ushort *addr;
	image_header_t *hdr;
	unsigned long long tmp;

	/* determine load address */
	if (argc > 2) {

		load_addr = simple_strtoul(argv[2], NULL, 16);

	} else if ((s = getenv("loadaddr")) != NULL) {

		load_addr = simple_strtoul(s, NULL, 16);

	}

	hdr = (image_header_t *)addr = (ushort *)load_addr;

	/* load image header from bank 1 */
	for (i = 0; i < sizeof(image_header_t); i += sizeof(ushort)) {

		read_via_fpu((unsigned long long *)
		             (FLASH_BANK1_BASE + (i << 2)), &tmp);

		*(addr++) = (ushort)((tmp & 0xffff000000000000) >> 48);
		
	}

	/* verify image header */
	if (hdr->ih_magic  != IH_MAGIC) {
		printf ("Bad Magic Number\n");
		return 1;
	}

	checksum = hdr->ih_hcrc;
	hdr->ih_hcrc = 0;

	if (crc32(0, (char *)hdr, sizeof(image_header_t)) != checksum) {
		printf ("Bad Header Checksum\n");
		return 1;
	}

	/* Restore checksum value... */
	hdr->ih_hcrc = checksum;

	if ((sizeof(image_header_t) + hdr->ih_size) > 0x00200000) {
		printf("Image overflows bank 1!\n");
		return 1;
	}

	printf ("   Loading Image:   %.*s\n", IH_NMLEN, hdr->ih_name);

	/* read rest of image */
	for (i = ((sizeof(image_header_t) + 1) & ~(0x1));
	     i < (hdr->ih_size + sizeof(image_header_t));
	     i += sizeof(ushort)) {

		read_via_fpu((unsigned long long *)
		             (FLASH_BANK1_BASE + (i << 2)), &tmp);

		*(addr++) = (ushort)((tmp & 0xffff000000000000) >> 48);
		
	}

	/* verify image data */
	printf("   Verifying Checksum ... ");
	if (crc32(0, (char *)(hdr + 1), hdr->ih_size) != hdr->ih_dcrc) {
		printf("Bad Data CRC\n");
		return 1;
	}
	printf("OK\n");

	return 0;
}

static int flbank1_save(cmd_tbl_t *cmdtp, int argc, char *argv[])
{
	char *s;
	int i;
	int j;
	int poll_count;
	ulong checksum;
	ulong erased = 0;
	ushort *addr;
	image_header_t *hdr;
	unsigned long long tmp;

	/* determine "load" address */
	if (argc > 2) {

		load_addr = simple_strtoul(argv[2], NULL, 16);

	} else if ((s = getenv("loadaddr")) != NULL) {

		load_addr = simple_strtoul(s, NULL, 16);

	}

	hdr = (image_header_t *)addr = (ushort *)load_addr;

	/* verify image header */
	if (hdr->ih_magic  != IH_MAGIC) {
		printf ("Bad Magic Number\n");
		return 1;
	}

	checksum = hdr->ih_hcrc;
	hdr->ih_hcrc = 0;

	if (crc32(0, (char *)hdr, sizeof(image_header_t)) != checksum) {
		printf ("Bad Header Checksum\n");
		return 1;
	}

	/* Restore checksum value... */
	hdr->ih_hcrc = checksum;

	if ((sizeof(image_header_t) + hdr->ih_size) > 0x00200000) {
		printf("Image is too big for bank 1!\n");
		return 1;
	}

	/* verify image data */
	printf("   Verifying Checksum ... ");
	if (crc32(0, (char *)(hdr + 1), hdr->ih_size) != hdr->ih_dcrc) {
		printf("Bad Data CRC\n");
		return 1;
	}
	printf("OK\n");

	/* Erase enough sectors... */
	printf ("   Erasing Sectors ...");
	for (i = 0;
	     (i < NB_OF(sinfo))
	  && (erased < (sizeof(image_header_t) + hdr->ih_size));
	     i++) {

		for (j = 0;
		     (j < sinfo[i].num_sects)
		  && (erased < (sizeof(image_header_t) + hdr->ih_size));
		     j++) {

			tmp = 0x00aa000000000000;
			write_via_fpu((unsigned long long *)
			              (FLASH_BANK1_BASE + (0x0aaa << 2)), &tmp);

			tmp = 0x0055000000000000;
			write_via_fpu((unsigned long long *)
			              (FLASH_BANK1_BASE + (0x0555 << 2)), &tmp);

			tmp = 0x0080000000000000;
			write_via_fpu((unsigned long long *)
			              (FLASH_BANK1_BASE + (0x0aaa << 2)), &tmp);

			tmp = 0x00aa000000000000;
			write_via_fpu((unsigned long long *)
			              (FLASH_BANK1_BASE + (0x0aaa << 2)), &tmp);

			tmp = 0x0055000000000000;
			write_via_fpu((unsigned long long *)
			              (FLASH_BANK1_BASE + (0x0555 << 2)), &tmp);

			tmp = 0x0030000000000000;
			write_via_fpu((unsigned long long *)
			              (FLASH_BANK1_BASE + (erased << 2)), &tmp);

			for (poll_count = 0;
			     poll_count < FLASH_BANK1_ERASE_TIMEOUT;
			     poll_count++) {

				udelay(FLASH_BANK1_POLL_DELAY);

				read_via_fpu((unsigned long long *)
				             (FLASH_BANK1_BASE + (erased << 2)),
				             &tmp);

				if ((tmp & 0xffff000000000000)
				 == 0xffff000000000000) break;

			}

			if (poll_count >= FLASH_BANK1_ERASE_TIMEOUT) {

				printf("Flash erase failed @ 0x%08lx\n",
				       (FLASH_BANK1_BASE + (erased << 2)));
				return 1;

			}

			printf (".");
			erased += sinfo[i].sect_size;

		}

	}
	printf(" Done.\n");

	printf ("   Saving Image:   %.*s\n", IH_NMLEN, hdr->ih_name);

	/* save image to bank 1 */
	for (i = 0;
	     i < (sizeof(image_header_t) + hdr->ih_size);
	     i += sizeof(ushort)) {

		tmp = 0x00aa000000000000;
		write_via_fpu((unsigned long long *)
		              (FLASH_BANK1_BASE + (0x0aaa << 2)), &tmp);

		tmp = 0x0055000000000000;
		write_via_fpu((unsigned long long *)
		              (FLASH_BANK1_BASE + (0x0555 << 2)), &tmp);

		tmp = 0x00a0000000000000;
		write_via_fpu((unsigned long long *)
		              (FLASH_BANK1_BASE + (0x0aaa << 2)), &tmp);

		tmp = ((unsigned long long)*(addr++) << 48);

		write_via_fpu((unsigned long long *)
		              (FLASH_BANK1_BASE + (i << 2)), &tmp);

		for (poll_count = 0;
		     poll_count < FLASH_BANK1_WRITE_TIMEOUT;
		     poll_count++) {

			read_via_fpu((unsigned long long *)
			             (FLASH_BANK1_BASE + (i << 2)),
			             &tmp);

			if ((tmp & 0xffff000000000000)
			 == ((unsigned long long)*(addr - 1) << 48)) 
				break;

			udelay(FLASH_BANK1_POLL_DELAY);

		}

		if (poll_count >= FLASH_BANK1_WRITE_TIMEOUT) {

			printf("Flash write failed @ 0x%08x\n",
			       (FLASH_BANK1_BASE + (i << 2)));
			return 1;

		}

	}

	return 0;
}

int do_flbank1(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
	if (argc > 1) {

		if (!strcmp(argv[1], "load")) {

			return (flbank1_load(cmdtp, argc, argv));

		} else if (!strcmp(argv[1], "save")) {

			return (flbank1_save(cmdtp, argc, argv));

		}
	}

	printf("Usage:\n%s\n", cmdtp->usage);
	return 1;
}

#endif /* (CONFIG_COMMANDS & CFG_CMD_BSP) */
